"""Load parameters for all scripts."""

# mis apres examination des donnees, on ignore les 10 premieres secondes (donnees samplee a 200Hz)
OFFSET = 2000
# mis en fonction de Van puten qui utilise des trials de 2s samples a 200Hz (utilise les donnees ds200)
NBINS = 129


def group1(age):
    if 18 <= age <= 28:
        return 0
    elif 28 <= age <= 38:
        return 1
    elif 38 <= age <= 48:
        return 2
    elif 48 <= age <= 58:
        return 3
    elif 58 <= age <= 68:
        return 4
    elif 68 <= age <= 78:
        return 5
    elif 78 <= age <= 88:
        return 6


FREQ_DICT = {
    "Delta": (1, 4),
    "Theta": (4, 8),
    "Alpha": (8, 15),
    "Beta": (15, 30),
    "Gamma1": (30, 60),
    "Gamma2": (60, 90),
    "Gamma3": (90, 120),
}

SF = 1000  # sampling frequency
N_ELEC = 306  # number electrodes

# Pour calculs PSD
WINDOW = 3000  # fenetre
OVERLAP = 0.25  # overlap (entre 0 et 1)

# Commenting this whole part as it is no longer necessary to have those here.
# keeping it in case we want to go back.

# SUB_DF = pd.read_csv("./clean_participant_new.csv", index_col=0)
# SUBJECT_LIST = SUB_DF["participant_id"].tolist()
#
# CHAN_DF = pd.read_csv("./channel_names.csv", index_col=0)
# CHANNEL_NAMES = CHAN_DF["ch_name"].tolist()
#
# AGES = SUB_DF["age"].tolist()
#
# LABELS = {}
# LABELS["sex"] = SUB_DF["sex"].tolist()
# LABELS["age"] = [group1(i) for i in AGES]
# LABELS["age_all"] = AGES
# LABELS["subject"] = list(range(len(SUBJECT_LIST)))



import os
from collections import defaultdict, OrderedDict
import logging
from typing import Tuple
import torch
from torch import nn, optim
from torch.autograd import Variable
from torch.utils.data import DataLoader
from scipy.io import loadmat, savemat
from matplotlib import pyplot as plt
import numpy as np
from huggingface_hub import hf_hub_download
from meegnet.layer import Flatten, DepthwiseConv2d, SeparableConv2d, Conv2dWithConstraint

LOG = logging.getLogger("meegnet")


def create_net(
    net_option: str, input_size: int, n_outputs: int, net_params: dict = None
) -> nn.Module:
    """
    Creates a neural network based on the specified option.

    Parameters
    ----------
    net_option : str
        Network architecture option ("MLP", "meegnet", "custom", "VGG", "EEGNet", "vanPutNet").
    input_size : int
        Input size of the network.
    n_outputs : int
        Number of output neurons.
    net_params : dict, optional
        Network parameters (required for "MLP" and "custom" options).

    Returns
    -------
    nn.Module
        Created neural network.

    Raises
    ------
    AttributeError
        If net_option is invalid.
    """

    if net_params is None:
        net_params = {
            "linear": 100,
            "hlayers": 2,
            "dropout": 0.5,
        }

    net_options = {
        "mlp": lambda: MLP(
            input_size=input_size,
            n_outputs=n_outputs,
            hparams={
                "mlp_width": net_params["linear"],
                "mlp_depth": net_params["hlayers"],
                "mlp_dropout": net_params["dropout"],
            },
        ),
        "meegnet": lambda: MEEGNet(input_size, n_outputs),
        "custom": lambda: FullNet(
            input_size,
            n_outputs,
            net_params["hlayers"],
            net_params["filters"],
            net_params["nchan"],
            net_params["linear"],
            net_params["dropout"],
            net_params["batchnorm"],
            net_params["maxpool"],
        ),
        "vgg": lambda: VGG16(input_size, n_outputs),
        "eegnet": lambda: EEGNet(input_size, n_outputs),
        # "eegnet": lambda: EEGNetv4(input_size, n_outputs),
        "vanputnet": lambda: VanPutNet(input_size, n_outputs),
    }

    if net_option.lower() not in net_options:
        raise AttributeError(f"Invalid network option: {net_option}")

    if net_option.lower() in ["mlp", "custom"] and net_params is None:
        raise ValueError("net_params is required for MLP and custom networks")

    return net_options[net_option.lower()]()


class CustomNet(nn.Module):
    """
    Base class for custom neural networks.

    Parameters
    ----------
    input_size : tuple
        Input shape of the network.
    n_outputs : int
        Number of output neurons.

    Attributes
    ----------
    input_size : tuple
        Input shape of the network.
    n_outputs : int
        Number of output neurons.

    Methods
    -------
    forward(x)
        Defines the forward pass through the network.
    get_output_shape(layers)
        Computes the output shape of a sequence of layers.
    feature_extraction(x)
        Defines the feature extraction part of the network (must be implemented).
    classif(x)
        Defines the classification part of the network (must be implemented).
    """

    def __init__(self, input_size: tuple, n_outputs: int) -> None:
        """
        Initializes the CustomNet.

        Args:
        input_size (tuple): Input shape of the network.
        n_outputs (int): Number of output neurons.
        """
        super().__init__()
        self.input_size = input_size
        self.n_outputs = n_outputs

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        """
        Defines the forward pass through the network.

        Args:
        x (torch.Tensor): Input tensor.

        Returns:
        torch.Tensor: Output tensor.
        """
        feats = self.feature_extraction(x)
        outs = self.classif(feats)
        return outs

    def get_output_shape(self, layers: nn.Sequential = None) -> tuple:
        if layers is None:
            layers = self.feature_extraction
        return layers(torch.zeros((1, *self.input_size))).shape

    def get_lin_size(self, layers: nn.Sequential) -> int:
        """
        Computes the output size of a sequence of layers.

        Args:
        layers (nn.Sequential): Sequence of layers.

        Returns:
        int: Output size of the sequence.
        """
        return self.get_output_shape(layers)[-1]


class EEGNetv4(CustomNet):
    def __init__(
        self,
        input_size: tuple,
        n_outputs: int,
        final_conv_length: str | int = "auto",
        pool_mode: str = "mean",
        F1: int = 8,
        D: int = 2,
        F2: int | None = None,
        kernel_length: int = 64,
        depthwise_kernel_length: int = 16,
        pool1_kernel_size: int = 4,
        pool1_stride_size: int = 4,
        pool2_kernel_size: int = 8,
        pool2_stride_size: int = 8,
        conv_spatial_max_norm: float = 1,
        activation: nn.Module = nn.ELU,
        batch_norm_momentum: float = 0.01,
        batch_norm_affine: bool = True,
        batch_norm_eps: float = 1e-3,
        drop_prob: float = 0.25,
    ) -> None:
        super().__init__(input_size, n_outputs)

        self.final_conv_length = final_conv_length
        self.pool_mode = pool_mode
        self.F1 = F1
        self.D = D
        self.F2 = F2 if F2 is not None else self.F1 * D
        self.kernel_length = kernel_length
        self.depthwise_kernel_length = depthwise_kernel_length
        self.pool1_kernel_size = pool1_kernel_size
        self.pool1_stride_size = pool1_stride_size
        self.pool2_kernel_size = pool2_kernel_size
        self.pool2_stride_size = pool2_stride_size
        self.conv_spatial_max_norm = conv_spatial_max_norm
        self.activation = activation
        self.batch_norm_momentum = batch_norm_momentum
        self.batch_norm_affine = batch_norm_affine
        self.batch_norm_eps = batch_norm_eps
        self.drop_prob = drop_prob

        pool_class = {"mean": nn.AvgPool2d, "max": nn.MaxPool2d}

        self.feature_extraction = nn.Sequential(
            # Rearrange("batch ch t -> batch 1 ch t"),
            nn.Conv2d(
                input_size[0],
                F1,
                (1, kernel_length),
                stride=1,
                bias=False,
                padding=(0, kernel_length // 2),
            ),
            nn.BatchNorm2d(
                self.F1,
                momentum=batch_norm_momentum,
                affine=batch_norm_affine,
                eps=batch_norm_eps,
            ),
            Conv2dWithConstraint(
                self.F1,
                self.F1 * self.D,
                (input_size[0], 1),
                max_norm=conv_spatial_max_norm,
                stride=1,
                bias=False,
                groups=self.F1,
                padding=(0, 0),
            ),
            nn.BatchNorm2d(
                self.F1 * self.D,
                momentum=batch_norm_momentum,
                affine=batch_norm_affine,
                eps=batch_norm_eps,
            ),
            activation(),
            pool_class[pool_mode](
                kernel_size=(1, pool1_kernel_size),
                stride=(1, pool1_stride_size),
            ),
            nn.Dropout(p=drop_prob),
            nn.Conv2d(
                self.F1 * self.D,
                self.F1 * self.D,
                (1, depthwise_kernel_length),
                stride=1,
                bias=False,
                groups=self.F1 * self.D,
                padding=(0, depthwise_kernel_length // 2),
            ),
            nn.Conv2d(
                self.F1 * self.D,
                self.F2,
                (1, 1),
                stride=1,
                bias=False,
                padding=(0, 0),
            ),
            nn.BatchNorm2d(
                self.F2,
                momentum=batch_norm_momentum,
                affine=batch_norm_affine,
                eps=batch_norm_eps,
            ),
            activation(),
            pool_class[pool_mode](
                kernel_size=(1, pool2_kernel_size),
                stride=(1, pool2_stride_size),
            ),
            nn.Dropout(p=drop_prob),
        )

        output_shape = self.get_output_shape()
        n_out_virtual_chans = output_shape[2]

        if self.final_conv_length == "auto":
            n_out_time = output_shape[3]
            self.final_conv_length = n_out_time

        self.classif = nn.Sequential(
            nn.Conv2d(
                self.F2,
                n_outputs,
                (n_out_virtual_chans, self.final_conv_length),
                bias=True,
            ),
            # Rearrange("batch x y z -> batch x z y"),
        )


class EEGNet(CustomNet):
    """
    EEGNet architecture implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape (S, C, T) where S is the number of sensors.
    n_outputs : int
        Number of output classes.
    filter_size : int, optional
        Filter size. Defaults to 64.
    n_filters : int, optional
        Number of filters. Defaults to 16.
    dropout : float, optional
        Dropout rate. Defaults to 0.5.
    dropout_option : str, optional
        Dropout type ("SpatialDropout2D" or "Dropout"). Defaults to "Dropout".
    depthwise_multiplier : int, optional
        Depthwise multiplier. Defaults to 2.

    Attributes
    ----------
    feature_extraction : nn.Sequential
        Feature extraction layers.
    classif : nn.Sequential
        Classification layers.
    """

    def __init__(
        self,
        input_size: tuple,
        n_outputs: int,
        filter_size: int = 64,
        n_filters: int = 16,
        dropout: float = 0.5,
        dropout_option: str = "Dropout",
        depthwise_multiplier: int = 2,
    ) -> None:
        super().__init__(input_size, n_outputs)

        if dropout_option not in ["SpatialDropout2D", "Dropout"]:
            raise ValueError(
                "dropout_option must be one of SpatialDropout2D "
                "or Dropout, passed as a string."
            )
        dropout_type = nn.Dropout2d if dropout_option == "SpatialDropout2D" else nn.Dropout

        n_channels = input_size[1]
        self.feature_extraction = nn.Sequential(
            nn.Conv2d(
                input_size[0],
                n_filters,
                (1, filter_size),
                padding=(1, filter_size // 2),
                bias=False,
            ),
            nn.BatchNorm2d(n_filters),
            DepthwiseConv2d(
                n_filters,
                (n_channels, 1),
                depthwise_multiplier=depthwise_multiplier,
                padding=0,
                bias=False,
            ),
            nn.BatchNorm2d(n_filters * depthwise_multiplier),
            nn.ELU(),
            nn.AvgPool2d((1, 4)),
            dropout_type(dropout),
            SeparableConv2d(
                n_filters * depthwise_multiplier,
                n_filters * depthwise_multiplier,
                (1, 16),
                padding=(1, 8),
                bias=False,
            ),
            nn.BatchNorm2d(n_filters * depthwise_multiplier),
            nn.ELU(),
            nn.AvgPool2d((1, 8)),
            dropout_type(dropout),
            Flatten(),
        )

        lin_size = self.get_lin_size(self.feature_extraction)
        self.classif = nn.Sequential(nn.Linear(lin_size, n_outputs))


# This implementation is rather common and found on various blogs/github repos
class VGG16(CustomNet):
    """
    VGG16 architecture implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape (C, H, W) or (S, C, T) for M/EEG data. With S the number of sensors,
        C the number of channels and T the number of time samples
    n_outputs : int
        Number of output classes.
    """

    def __init__(self, input_size: tuple, n_outputs: int) -> None:
        super().__init__(input_size, n_outputs)

        in_channels = input_size[0]
        self.feature_extraction = nn.Sequential(
            nn.Conv2d(in_channels, 64, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(64, 64, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(64, 128, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(128, 128, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(128, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(256, 256, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            nn.Conv2d(256, 512, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(512, 512, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.Conv2d(512, 512, kernel_size=3, padding=1),
            nn.ReLU(),
            nn.MaxPool2d(kernel_size=2, stride=2),
            Flatten(),
        )

        lin_size = self.get_lin_size(self.feature_extraction)
        self.classif = nn.Sequential(
            nn.Linear(lin_size, 4096),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(4096, 4096),
            nn.ReLU(),
            nn.Dropout(),
            nn.Linear(4096, n_outputs),
        )


class MLP(CustomNet):
    """
    Multi-Layer Perceptron (MLP) implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape.
    n_outputs : int
        Number of output neurons.
    hparams : dict
        Hyperparameters:
            - mlp_width (int): Number of neurons in each hidden layer.
            - mlp_depth (int): Number of hidden layers.
            - mlp_dropout (float): Dropout rate.
    """

    def __init__(self, input_size: tuple, n_outputs: int, hparams: dict) -> None:
        super().__init__(input_size, n_outputs)
        n_inputs = np.prod(input_size)

        self.feature_extraction = nn.Sequential(
            Flatten(),
            nn.Linear(n_inputs, hparams["mlp_width"]),
            nn.Dropout(hparams["mlp_dropout"]),
            nn.ReLU(),
            *[
                nn.Sequential(
                    nn.Linear(hparams["mlp_width"], hparams["mlp_width"]),
                    nn.Dropout(hparams["mlp_dropout"]),
                    nn.ReLU(),
                )
                for _ in range(hparams["mlp_depth"] - 2)
            ],
        )

        self.classif = nn.Linear(hparams["mlp_width"], n_outputs)


class MEEGNet(CustomNet):
    """
    MEEGNet architecture implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape (S, C, T) where S is the number of sensors.
    n_outputs : int
        Number of output classes.
    n_linear : int, optional
        Number of neurons in the first linear layer. Defaults to 2000.
    dropout : float, optional
        Dropout rate. Defaults to 0.5.
    """

    def __init__(
        self,
        input_size: tuple,
        n_outputs: int,
        n_linear: int = 2000,
        dropout: float = 0.5,
    ) -> None:
        super().__init__(input_size, n_outputs)

        maxpool = nn.MaxPool2d(kernel_size=(1, 20), stride=1)

        self.feature_extraction = nn.Sequential(
            nn.Conv2d(input_size[0], 100, (input_size[1], 1)),
            nn.ReLU(),
            nn.Conv2d(100, 200, (1, 9)),
            maxpool,
            nn.ReLU(),
            nn.Conv2d(200, 200, (1, 9)),
            maxpool,
            nn.ReLU(),
            nn.Conv2d(200, 100, (1, 9)),
            maxpool,
            nn.ReLU(),
            Flatten(),
            nn.Dropout(dropout),
        )

        lin_size = self.get_lin_size(self.feature_extraction)
        self.classif = nn.Sequential(
            nn.Linear(lin_size, n_linear // 2),
            nn.Linear(n_linear // 2, n_outputs),
        )


class FullNet(CustomNet):
    """
    FullNet architecture implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape (S, C, T) where S is the number of sensors.
    n_outputs : int
        Number of output classes.
    hlayers : int, optional
        Number of hidden layers. Defaults to 2.
    filter_size : int, optional
        Filter size. Defaults to 7.
    nchan : int, optional
        Number of channels. Defaults to 5.
    n_linear : int, optional
        Number of neurons in the first linear layer. Defaults to 150.
    dropout : float, optional
        Dropout rate. Defaults to 0.25.
    batchnorm : bool, optional
        Whether to use batch normalization. Defaults to False.
    maxpool : int, optional
        Max pooling size. Defaults to 0.
    """

    def __init__(
        self,
        input_size: tuple,
        n_outputs: int,
        hlayers: int = 2,
        filter_size: int = 7,
        nchan: int = 5,
        n_linear: int = 150,
        dropout: float = 0.25,
        batchnorm: bool = False,
        maxpool: int = 0,
    ) -> None:
        super().__init__()

        self.feature_extraction = nn.Sequential(
            nn.Conv2d(input_size[0], nchan, (input_size[1], 1)),
            nn.ReLU(),
            *self._build_conv_layers(hlayers, filter_size, nchan, batchnorm, maxpool),
            Flatten(),
            nn.Dropout(dropout),
        )

        lin_size = self.get_lin_size(self.feature_extraction)
        self.classif = nn.Sequential(
            nn.Linear(lin_size, n_linear // 2),
            nn.Linear(n_linear // 2, n_outputs),
        )

    def _build_conv_layers(self, hlayers, filter_size, nchan, batchnorm, maxpool):
        layers = []
        prev = nchan

        # Encoder
        for i in range(hlayers // 2):
            nex = prev * 2
            layers += [
                nn.Conv2d(prev, nex, (1, filter_size)),
                nn.ReLU(),
                *([nn.BatchNorm2d(nex)] if batchnorm else []),
                *([nn.MaxPool2d((1, maxpool), 1)] if maxpool != 0 else []),
            ]
            prev = nex

        # Middle layer (if odd number of hidden layers)
        if hlayers % 2 != 0:
            layers += [
                nn.Conv2d(prev, prev, (1, filter_size)),
                nn.ReLU(),
                *([nn.BatchNorm2d(prev)] if batchnorm else []),
                *([nn.MaxPool2d((1, maxpool), 1)] if maxpool != 0 else []),
            ]

        # Decoder
        for i in range(hlayers // 2):
            nex = prev // 2
            layers += [
                nn.Conv2d(prev, nex, (1, filter_size)),
                nn.ReLU(),
                *([nn.BatchNorm2d(nex)] if batchnorm else []),
                *([nn.MaxPool2d((1, maxpool), 1)] if maxpool != 0 else []),
            ]
            prev = nex

        return layers


class VanPutNet(CustomNet):
    """
    VanPutNet architecture implementation.

    Parameters
    ----------
    input_size : tuple
        Input shape (C, H, W) or (S, C, T) for M/EEG data.
    n_output : int
        Number of output classes.
    dropout : float, optional
        Dropout rate. Defaults to 0.25.
    """

    def __init__(self, input_size: tuple, n_output: int, dropout: float = 0.25) -> None:
        super().__init__(input_size, n_output)

        in_channels = input_size[0]
        self.feature_extraction = nn.Sequential(
            nn.Conv2d(in_channels, 100, 3),
            nn.ReLU(),
            nn.MaxPool2d((2, 2)),
            nn.Dropout(dropout),
            nn.Conv2d(100, 100, 3),
            nn.MaxPool2d((2, 2)),
            nn.Dropout(dropout),
            nn.Conv2d(100, 300, (2, 3)),
            nn.MaxPool2d((2, 2)),
            nn.Dropout(dropout),
            nn.Conv2d(300, 300, (1, 7)),
            nn.MaxPool2d((1, 2)),
            nn.Dropout(dropout),
            nn.Conv2d(300, 100, (1, 3)),
            nn.Conv2d(100, 100, (1, 3)),
            Flatten(),
        )

        lin_size = self.get_lin_size(self.feature_extraction)
        self.classif = nn.Sequential(
            nn.Linear(lin_size, n_output),
            nn.Softmax(dim=1),
        )


class AutoEncoder(CustomNet):
    def __init__(
        self,
        input_size,
    ):
        CustomNet.__init__(self, input_size)

        lin_size = input_size[0] * input_size[1] * input_size[2]

        self.encoder = nn.Sequential(
            nn.Linear(lin_size, 4096),
            nn.ReLU(True),
            nn.Linear(4096, 2048),
        )

        self.decoder = nn.Sequential(
            nn.Linear(2048, 4096),
            nn.ReLU(True),
            nn.Linear(4096, lin_size),
            nn.Tanh(),
        )

    def forward(self, x):
        x = self.encoder(x)
        return self.decoder(x)


class Model:
    def __init__(
        self,
        name: str,
        net_option: str,
        input_size: tuple,
        n_outputs: int,
        save_path: str = None,
        learning_rate: float = 0.00001,
        optimizer: callable = optim.Adam,
        criterion: callable = nn.CrossEntropyLoss(),
        n_folds: int = 5,
        device: str = "cuda",
        net_params: dict = None,
    ) -> None:
        """
        Initialize the Model class.

        Parameters
        ----------
        name : str
            Model name.
        net_option : str
            Network architecture option.
        input_size : tuple
            Input shape (C, H, W) or (S, C, T) for EEG data.
        n_outputs : int
            Number of output classes.
        save_path : str, optional
            Model save path. Defaults to None.
        learning_rate : float, optional
            Learning rate. Defaults to 0.00001.
        optimizer : callable, optional
            Optimizer function. Defaults to Adam.
        criterion : callable, optional
            Loss function. Defaults to CrossEntropyLoss.
        n_folds : int, optional
            Number of folds for cross-validation. Defaults to 5.
        device : str, optional
            Device to use (cuda or cpu). Defaults to "cuda".
        net_params : dict, optional
            Network architecture parameters. Defaults to None.
        """
        assert isinstance(name, str), "Name must be a string"
        assert isinstance(input_size, tuple), "Input size must be a tuple"
        assert len(input_size) == 3, "Input size must have 3 dimensions"

        self.name = name
        self.input_size = input_size  # TODO here put assertions on the shape
        self.net = create_net(net_option, input_size, n_outputs, net_params)
        self.n_outputs = n_outputs
        self.n_folds = n_folds
        self.criterion = criterion
        self.save_path = save_path
        self.lr = learning_rate
        self.optimizer = optimizer(self.net.parameters(), lr=learning_rate)
        self.tracker = TrainingTracker(self.save_path, self.name)

        if torch.cuda.is_available():
            self.device = "cuda"
        elif device == "cuda":
            LOG.warning("Warning: gpu device requested but unavailable. Setting device to CPU")
            self.device = "cpu"
        else:
            self.device = "cpu"
        self.net.to(self.device)

    def train(
        self,
        dataset,
        batch_size: int = 128,
        patience: int = 10,
        max_epoch: int = None,
        model_path: str = None,
        early_stop: str = "loss",
        num_workers: int = 4,
        continue_training: bool = False,
        verbose: int = 3,
    ) -> None:
        """
        Train the model on the provided dataset.

        Parameters
        ----------
        dataset
            Dataset to train on.
        batch_size : int, optional
            Batch size for training. Defaults to 128.
        patience : int, optional
            Patience for early stopping. Defaults to 10.
        max_epoch : int, optional
            Maximum number of epochs to train. Defaults to None.
        model_path : str, optional
            Path to save the model. Defaults to None.
        num_workers : int, optional
            Number of workers for data loading. Defaults to 4.
        continue_training : bool, optional
            Wether to pick up training from last checkpoint. By default False. Use when
            training stopped abbruptly because of an error, or to re-train or fine-tune
            the model.
        verbose : int, optional
            The verbosity level. By default 3.
            Values under 3 will display less detailed progress during training.
            Values under 2 will not display any update on performance epoch per epoch during training.

        Notes
        -----
        This method trains the model using the provided dataset and hyperparameters.
        It uses early stopping based on the validation loss and saves the model
        periodically.
        """
        assert len(dataset.data) > 0, "Dataset is empty."
        # Check dataset compatibility
        assert (
            dataset.data[0].shape == self.input_size
        ), "Dataset sample size must match network input size."
        assert early_stop in (
            "loss",
            "accuracy",
        ), f"{early_stop} is not a valid early_stop option."

        # Setting model_path
        self.tracker.set_model_path(model_path)

        # Set training mode and batch size
        self.net.train()
        self.batch_size = batch_size
        self.num_workers = num_workers

        # Create data loaders
        LOG.info("Creating DataLoaders...")
        train_index, valid_index, _ = dataset.split_data()
        trainloader = DataLoader(
            dataset.torchDataset(train_index),
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=True,
        )
        validloader = DataLoader(
            dataset.torchDataset(valid_index),
            batch_size=batch_size,
            num_workers=num_workers,
            shuffle=True,
            pin_memory=True,
        )

        # Log training configuration
        LOG.info("Starting Training with:")
        LOG.info(f"Batch size: {batch_size}")
        LOG.info(f"Learning rate: {self.lr}")
        LOG.info(f"Patience: {patience}")
        if max_epoch is not None:
            LOG.info(f"Maximum Epoch: {max_epoch}")

        epoch = 0
        if continue_training:
            epoch = self.tracker.best["epoch"] + 1
            self.tracker.patience_state = 0

        while self.tracker.patience_state < patience and (
            max_epoch is None or epoch < max_epoch
        ):
            self.train_epoch(epoch, trainloader, verbose=verbose)
            train_loss, train_acc = self.evaluate(trainloader)
            valid_loss, valid_acc = self.evaluate(validloader)
            self.tracker.update(
                epoch,
                train_loss,
                train_acc,
                valid_loss,
                valid_acc,
                self.net,
                self.optimizer,
                early_stop=early_stop,
            )
            if verbose >= 2:
                LOG.info(f"Epoch: {epoch}")
                LOG.info(f" [LOSS] TRAIN {train_loss:.4f} / VALID {valid_loss:.4f}")
                LOG.info(f" [ACC] TRAIN {100*train_acc:.2f}% / VALID {100*valid_acc:.2f}%")
            epoch += 1

    def fit(self, *args, **kwargs):
        return self.train(args, *kwargs)

    def train_batch(self, batch):
        self.optimizer.zero_grad()
        if len(batch) > 2:
            X, y, groups = batch
        else:
            X, y = batch
        y = y.view(-1)
        targets = Variable(y.type(torch.LongTensor)).to(self.device)
        X = X.view(-1, *self.input_size).to(self.device)
        loss = self.criterion(self.net.forward(X), targets)
        loss = loss
        loss.backward()
        self.optimizer.step()
        return loss

    def display_progress(self, i, epoch, loss, n_batches, verbose=3):
        progress = f"Epoch: {epoch} // Batch {i+1}/{n_batches} // loss = {loss:.5f}"
        if n_batches > 10:
            if i % (n_batches // 10) == 0 and verbose > 2:
                LOG.info(progress)
        elif verbose > 2:
            LOG.info(progress)

    def train_epoch(self, epoch, trainloader, verbose):
        # Train loop for a single epoch
        n_batches = len(trainloader)
        for i, batch in enumerate(trainloader):
            loss = self.train_batch(batch)
            self.display_progress(i, epoch, loss, n_batches, verbose=verbose)

    def evaluate(self, dataloader):
        with torch.no_grad():
            losses = 0
            accuracy = 0
            counter = 0
            for batch in dataloader:
                if len(batch) > 2:
                    X, y, _ = batch
                else:
                    X, y = batch
                y = y.view(-1)
                targets = Variable(y.type(torch.LongTensor)).to(self.device)
                X = X.view(-1, *self.input_size).to(self.device)
                out = self.net.forward(X)
                loss = self.criterion(out, targets)
                acc = self.compute_accuracy(out, targets)
                n = y.size(0)
                losses += loss.detach().sum().data.cpu().numpy() * n
                accuracy += acc.sum().data.cpu().numpy() * n
                counter += n
            floss = losses / float(counter)
            faccuracy = accuracy / float(counter)
            return floss, faccuracy

    def test(self, dataset):
        _, _, test_index = dataset.split_data()
        test_loader = DataLoader(
            dataset.torchDataset(test_index),
            batch_size=self.batch_size,
            num_workers=self.num_workers,
            shuffle=True,
            pin_memory=True,
        )
        test_loss, test_acc = self.evaluate(test_loader)
        LOG.info(f" [LOSS] TEST {test_loss}")
        LOG.info(f" [ACC] TEST {test_acc}")
        return test_loss, test_acc

    def n_parameters(self, model: nn.Module = None):
        if model is None:
            model = self.net
        model_parameters = filter(lambda p: p.requires_grad, model.parameters())
        return sum([np.prod(p.size()) for p in model_parameters])

    def _get_from_hub(self, repo=None):
        if repo is None:
            repo = "lamaroufle/meegnet"
        model_name = "_".join(self.name.split("_")[:-2])
        input_size_string = "_".join(map(str, self.input_size))
        filename = f"{model_name}_{input_size_string}_{self.n_outputs}"
        model_path = hf_hub_download(repo_id="lamaroufle/meegnet", filename=filename + ".pt")
        hf_hub_download(repo_id="lamaroufle/meegnet", filename=filename + ".mat")
        return model_path

    def from_pretrained(self, repo=None):
        model_path = self._get_from_hub(repo)
        self.load(model_path)

    def _load_net(self, model_path: str = None) -> Tuple:
        """Load network state and optimizer state from file."""
        if model_path is None:
            model_path = self.tracker.model_path
        if os.path.exists(model_path):
            LOG.info("=> loading checkpoint '{}'".format(model_path))
            checkpoint = torch.load(model_path)
            net_state = checkpoint["state_dict"]
            optimizer_state = checkpoint["optimizer"]
            # net_state = checkpoint["state_dict"]()
            # optimizer_state = checkpoint["optimizer"].state_dict()
        mat_path = model_path[:-2] + "mat"
        if os.path.exists(mat_path):
            self.tracker.load(mat_path)
        else:
            LOG.warning(f"Error while loading checkpoint from {model_path}")
        return net_state, optimizer_state

    def load(self, model_path=None):
        """Load model from file."""
        try:
            net_state, optimizer_state = self._load_net(model_path)
            if net_state[list(net_state.keys())[-1]].shape[0] == self.n_outputs:
                self.net.load_state_dict(net_state)
                self.optimizer.load_state_dict(optimizer_state)
            else:
                feat_state_dict = OrderedDict()
                for key, value in net_state.items():
                    if key.startswith("feature"):
                        feat_state_dict[".".join(key.split(".")[1:])] = value
                self.net.feature_extraction.load_state_dict(feat_state_dict)
                self.tracker = TrainingTracker(self.save_path, self.name)
        except FileNotFoundError:
            LOG.error(f"Model file not found: {model_path}")

    def compute_accuracy(self, y_pred, target):
        # Compute
        correct = torch.eq(y_pred.max(1)[1], target).sum().type(torch.FloatTensor)
        return correct / len(target)

    def get_feature_weights(self):
        weights = []
        for layer in self.net.feature_extraction:
            if hasattr(layer, "weight"):
                weights.append(layer.weight.cpu().detach().numpy())
        return weights

    def get_clf_weights(self):
        weights = []
        for layer in self.net.classif:
            if hasattr(layer, "weight"):
                weights.append(layer.weight.detach().numpy())
        return weights

    def plot_accuracy(self, option="both"):
        return self.tracker.plot_accuracy(option)

    def plot_loss(self, option="both"):
        return self.tracker.plot_loss(option)

    def save(self, model_path: str = None):
        self.tracker.set_model_path(model_path)
        checkpoint = {
            "state_dict": self.net.state_dict(),
            "optimizer": self.optimizer.state_dict(),
        }
        self.tracker.save(checkpoint)


class TrainingTracker:
    def __init__(self, save_path, name, model_path: str = None):
        self.progress = {
            "train_losses": [],
            "train_accuracies": [],
            "validation_losses": [],
            "validation_accuracies": [],
        }
        self.best = {
            "train_loss": float("inf"),
            "train_accuracy": 0,
            "validation_loss": float("inf"),
            "validation_accuracy": 0,
            "epoch": 0,
        }
        self.patience_state = 0
        self.save_path = save_path
        self.name = name

        self.set_model_path(model_path)

    def set_model_path(self, model_path: str = None) -> None:
        self.model_path = (
            os.path.join(self.save_path, self.name + ".pt")
            if model_path is None
            else model_path
        )

    def update(
        self, epoch, tloss, tacc, vloss, vacc, net, optimizer, early_stop: str = "loss"
    ) -> None:
        assert early_stop in (
            "loss",
            "accuracy",
        ), f"{early_stop} is not a valid early_stop option."

        if epoch < len(self.progress["train_losses"]):
            self.progress["train_losses"][epoch] = tloss
            self.progress["train_accuracies"][epoch] = tacc
            self.progress["validation_losses"][epoch] = vloss
            self.progress["validation_accuracies"][epoch] = vacc
            self.patience_state += 1
        else:
            self.progress["train_losses"].append(tloss)
            self.progress["train_accuracies"].append(tacc)
            self.progress["validation_losses"].append(vloss)
            self.progress["validation_accuracies"].append(vacc)
            self.patience_state += 1

        check = {
            "loss": vloss < self.best["validation_loss"],
            "accuracy": vacc > self.best["validation_accuracy"],
        }
        if check[early_stop]:
            self.best["train_loss"] = tloss
            self.best["train_accuracy"] = tacc
            self.best["validation_loss"] = vloss
            self.best["validation_accuracy"] = vacc
            self.best["epoch"] = epoch
            self.patience_state = 0
            checkpoint = {"state_dict": net.state_dict(), "optimizer": optimizer.state_dict()}
            self.save(checkpoint)

    def save(self, checkpoint) -> None:
        """Save model to file."""
        mat_path = self.model_path[:-2] + "mat"
        try:
            torch.save(checkpoint, self.model_path)
            save_dict = {key: value for key, value in self.progress.items()}
            save_dict.update({key: value for key, value in self.best.items()})
            savemat(mat_path, save_dict)
        except OSError:
            LOG.error(f"Error saving model to file: {self.model_path}")

    def load(self, mat_path):
        data = loadmat(mat_path)
        for key, value in data.items():
            if key in self.progress.keys():
                self.progress[key] = np.array(value).squeeze()
            elif key in self.best.keys():
                self.best[key] = np.array(value).squeeze()

    def plot_metric(self, metric_type: str, option: str = "both", early_stop: bool = True):
        assert option in ["both", "train", "valid"]
        assert metric_type in ["accuracy", "loss"]
        metric_key = {
            "accuracy": "accuracies",
            "loss": "losses",
        }

        fig, ax = plt.subplots()

        if option in ("both", "train"):
            plt.plot(
                np.array(self.progress[f"train_{metric_key[metric_type]}"]).squeeze(),
                label=f"Training {metric_type.capitalize()}",
            )
        if option in ("both", "valid"):
            plt.plot(
                np.array(self.progress[f"validation_{metric_key[metric_type]}"]).squeeze(),
                label=f"Validation {metric_type.capitalize()}",
            )

        epochs = len(self.progress[f"train_{metric_key[metric_type]}"])
        step = max(1, epochs // 10)  # Ensure at most 10 ticks
        ticks = list(range(0, epochs, step))  # Use epoch indices for ticks

        if early_stop:
            plt.axvline(x=self.best["epoch"], label="early stop", color="green")
            if self.best["epoch"] not in ticks:
                ticks.append(self.best["epoch"])

        ax.set_xticks(sorted(ticks))
        ax.set_xticklabels([x + 1 for x in sorted(ticks)])  # Labels should start from 1
        ax.set_ylabel(metric_type.capitalize())
        ax.set_xlabel("Epoch")
        plt.legend()
        plt.plot()
        return fig

    def plot_accuracy(self, option: str = "both", early_stop: bool = True):
        return self.plot_metric(metric_type="accuracy", option=option, early_stop=early_stop)

    def plot_loss(self, option: str = "both", early_stop: bool = True):
        return self.plot_metric(metric_type="loss", option=option, early_stop=early_stop)



import os
import random
import logging
import pandas as pd
import numpy as np
import torch
from torch.utils.data import random_split
from scipy.stats import zscore
from meegnet.utils import strip_string

LOG = logging.getLogger("meegnet")


class _InfiniteSampler(torch.utils.data.Sampler):
    """
    Wraps another Sampler to yield an infinite stream.

    Parameters
    ----------
    sampler : torch.utils.data.Sampler
        The sampler to be wrapped.
    """

    def __init__(self, sampler):
        self.sampler = sampler

    def __iter__(self):
        while True:
            for batch in self.sampler:
                yield batch


class InfiniteDataLoader:
    """
    Creates an infinite data loader.

    Parameters
    ----------
    dataset : torch.utils.data.Dataset
        The dataset to be loaded.
    batch_size : int
        The size of the batches to be loaded.
    num_workers : int, optional
        The number of workers to use for loading the data, by default 0.
    pin_memory : bool, optional
        If True, the data loader will copy tensors into CUDA pinned memory before returning them. This can make data
        transfer faster, but requires more memory.
    weights : torch.Tensor, optional
        A 1D tensor assigning a weight to each sample in the dataset. If not provided, all samples are assumed to have
        the same weight.

    Returns
    -------
    InfiniteDataLoader
        An infinite data loader.
    """

    def __init__(self, dataset, batch_size, num_workers=0, pin_memory=False, weights=None):
        super().__init__()

        if weights is not None:
            sampler = torch.utils.data.WeightedRandomSampler(
                weights, replacement=True, num_samples=batch_size
            )
        else:
            sampler = torch.utils.data.RandomSampler(dataset, replacement=True)

        if weights is None:
            weights = torch.ones(len(dataset))

        batch_sampler = torch.utils.data.BatchSampler(
            sampler, batch_size=batch_size, drop_last=True
        )

        self._infinite_iterator = iter(
            torch.utils.data.DataLoader(
                dataset,
                num_workers=num_workers,
                batch_sampler=_InfiniteSampler(batch_sampler),
                pin_memory=pin_memory,
            )
        )

    def __iter__(self):
        while True:
            yield next(self._infinite_iterator)

    def __len__(self):
        raise ValueError


def _string_to_int(array, target_labels=None):
    array = np.array(array)
    if target_labels is None:
        target_labels = np.unique(array)
    for i, element in enumerate(target_labels):
        array[array == element] = i
    return array.astype(int), target_labels


class EpochedDataset:
    """
    Creates a dataset for epoch-based M/EEG data.

    Parameters
    ----------
    sfreq : int, optional
        The sampling frequency, by default 500.
    n_subjects : int, optional
        The number of subjects. Default value is None, which means all subjects are processed.
    scaling : str, optional
        The scaling method. available options are "zscore" and "minmax". By default, minmax.
    n_samples : int, optional
        The number of samples to include, by default None.
    sensortype : str, optional
        The type of sensor to use, by default None.
    lso : bool, optional
        Leave subjects out. If False, within-subject splitting is used, by default False.
    random_state : int, optional
        The random state for reproducibility, by default 0.

    Attributes
    ----------
    sfreq : int
        The sampling frequency.
    n_subjects : int
        The number of subjects.
    n_samples : int
        The number of samples for each subject.
    data : torch.Tensor
        The data.
    targets : torch.Tensor
        The targets.
    groups : list
        The groups.
    subject_list : list
        The subject list.
    data_path : str
        The path to the data.
    """

    def __init__(
        self,
        sfreq: float = 500,
        n_subjects: int = None,
        scaling: str = "minmax",
        n_samples: int = None,
        split_sizes: tuple = (0.8, 0.1, 0.1),
        sensortype: str = None,
        lso: bool = False,
        random_state: int = 0,
        target_labels: list = None,
    ):
        if isinstance(split_sizes, float):
            split_sizes = split_sizes, (1 - split_sizes) / 2, (1 - split_sizes) / 2

        self._assert_sizes(*split_sizes)
        self.split_sizes = split_sizes

        self.sfreq = sfreq
        self.n_subjects = n_subjects
        if scaling == "zscore":
            self.scaler = lambda x: zscore(x, axis=-1)
        elif scaling == "minmax":
            self.scaler = lambda x: (x - x.min()) / (x.max() - x.min())
        else:
            raise ValueError(f"{scaling} is an invalid scaling option.")
        self.n_samples = n_samples
        self.lso = lso
        self.sensors = self._select_sensors(sensortype)
        self.random_state = random_state
        self.target_labels = target_labels
        self._reset_seed()

        self.data = []
        self.targets = []
        self.groups = []
        self.subject_list = []
        self.data_path = None
        self.dataframe = None

    def _reset_seed(self):
        np.random.seed(self.random_state)
        random.seed(self.random_state)
        torch.manual_seed(self.random_state)

    def _load_csv(self, csv_file: str) -> pd.DataFrame:
        """Loads a CSV file, handling index column."""
        with open(csv_file) as f:
            first_line = f.readline()
        return pd.read_csv(csv_file, index_col=0 if first_line.startswith(",") else None)

    def preload(
        self, data_path: str, csv_path: str = None, subject_col: str = "sub"
    ) -> pd.DataFrame:
        """
        Loads the subject list from a CSV file.

        Parameters
        ----------
        data_path : str
            Path to the folder containing the dataset.
        csv_path : str, optional
            Path to the CSV file. Defaults to "participants_info.csv" in the data_path.

        Returns
        -------
        pd.DataFrame
            DataFrame containing participant information.
        """

        # Determine CSV file path
        if csv_path is not None and os.path.exists(csv_path):
            dataframe = self._check_index_and_load_csv(csv_path)
        else:
            dataframe = self._load_csv(data_path)

        # Sample participants (if needed)
        if self.n_subjects < len(dataframe):
            self.dataframe = dataframe.sample(
                frac=1, random_state=self.random_state
            ).reset_index(drop=True)[: self.n_subjects]
        else:
            self.dataframe = dataframe

        # Update instance attributes
        LOG.info(f"Logging subjects and targets from {data_path}...")
        self.data_path = data_path
        self.subject_list = self.dataframe[subject_col].tolist()

        return self.dataframe

    def _check_index_and_load_csv(self, csv_file) -> pd.DataFrame:
        with open(csv_file) as f:
            first_line = f.readline()
        if first_line.startswith(","):
            df = pd.read_csv(csv_file, index_col=0)
        else:
            df = pd.read_csv(csv_file)
        return df

    def _load_csv(self, data_path, csv_name="participants_info.csv") -> pd.DataFrame:
        csv_file = os.path.join(data_path, csv_name)
        df = self._check_index_and_load_csv(csv_file)
        participants_df = df.sample(frac=1, random_state=self.random_state).reset_index(
            drop=True
        )[: self.n_subjects]
        return participants_df

    def load(
        self,
        data_path: str = None,
        csv_path: str = None,
        one_sub: str = None,
        verbose: int = 2,
        target_col: str = "label",
        subject_col: str = "sub",
    ) -> None:
        """
        Loads data from the "downsampled_[sfreq]" folder in the data_path.

        Parameters
        ----------
        data_path : str, optional
            Path to the data. Defaults to self.data_path if None.
        csv_path : str, optional
            Path to the CSV file. Defaults to "participants_info.csv" in data_path.
        one_sub : str, optional
            Subject ID or "random" to select a random subject.
        verbose : int, optional
            Logging verbosity level (0-2).
        target_col : str, optional
            Column name for targets in the CSV file.
        """

        # Ensure data_path is set
        if data_path is None:
            assert self.data_path is not None, "data_path must be set"
            data_path = self.data_path

        # Load participants info
        self.preload(data_path, csv_path, subject_col=subject_col)

        # Set logging verbosity
        verbosity_levels = {0: logging.NOTSET, 1: logging.WARNING, 2: logging.INFO}
        LOG.setLevel(verbosity_levels.get(verbose, logging.INFO))

        # Select subject(s) to load
        if one_sub == "random":
            one_sub = self.random_sub()
        elif one_sub is not None and one_sub not in self.subject_list:
            raise AttributeError(f"{one_sub} not a valid subject")
        elif one_sub is not None:
            self.subject_list = [one_sub]

        # Load data for selected subject(s)
        data_folder = f"downsampled_{self.sfreq}"
        numpy_filepath = os.path.join(self.data_path, data_folder)
        for file in os.listdir(numpy_filepath):
            self._reset_seed()
            sub = file.split("_")[0]  # The subject ID is placed first in the filename
            if one_sub is not None and sub != one_sub:
                continue
            if sub not in self.subject_list:
                continue

            row = self.dataframe.loc[self.dataframe[subject_col] == sub]
            sub_data = self._load_sub(os.path.join(numpy_filepath, file))
            if sub_data is None:
                continue  # skip subject if there are no data in the loaded file

            # Process data and targets
            if self.sensors is not None:
                sub_data = sub_data[:, self.sensors, :, :]
            target = row[target_col].item()
            targets = self._process_targets(target, len(sub_data))

            # Handle sampling
            if self.n_samples is not None:
                random_samples = np.random.choice(
                    np.arange(len(sub_data)), self.n_samples, replace=False
                )
                sub_data = sub_data[random_samples]
                targets = targets[random_samples]

            if len(sub_data) == len(targets):
                self.data.append(torch.Tensor(sub_data))
                self.targets.append(np.array(targets))
                self.groups += [sub] * len(targets)
            else:
                LOG.warning(
                    f"Warning: Number of trials for {sub} does not match number of targets."
                )
                continue

        # Format data and targets
        if len(self) == 0:
            return
        elif len(self) == 1:
            self.data = self.data[0]
            self.targets = self.targets[0]
        else:
            self.data = torch.cat(self.data, 0)
            self.targets = np.concatenate(self.targets, axis=0)

        if len(self.data.shape) != 4:
            self.data = self.data[:, np.newaxis]

        self.set_targets(self.targets, self.target_labels)

        if type(self.groups[0]) != int:
            self.groups, _ = _string_to_int(self.groups)
        self.groups = torch.tensor(self.groups, dtype=int)

        self.n_subjects = len(np.unique(self.groups))

    def _process_targets(self, target: str, n_samples: int) -> list:
        """Process target(s) for a subject."""
        if isinstance(target, str):
            targets = target.split(", ")
            targets = list(map(strip_string, targets))
        else:
            targets = [target]
        if len(targets) == 1:
            if targets[0] in self.subject_list:
                targets = [self.subject_list.index(targets[0])] * n_samples
            else:
                targets = [targets[0]] * n_samples
        return np.array(targets)

    def random_sub(self):
        return np.random.choice(self.subject_list)

    def __len__(self):
        """Returns the length of the dataset (total number of data examples)."""
        return len(self.data)

    def _load_sub(self, filepath: str):
        """Loads a single subject's data."""
        try:
            data = np.load(filepath)
        except IOError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        data = np.array([list(map(self.scaler, sensor_data)) for sensor_data in data])
        return torch.Tensor(np.array(data))

    def _select_sensors(self, sensortype: str) -> list:
        """
        Selects MEG sensor slices based on the sensor type.

        Parameters
        ----------
        sensortype : str
            Type of sensor ("MAG", "GRAD", "plannar", or variants).

        Returns
        -------
        list or None
            List of sensor indices if sensortype is valid, otherwise None.
        """
        sensor_mapping = {
            "MAG": [0],
            "GRAD": [1, 2],
            "plannar": [1, 2],
            "GRAD1": [1],
            "plannar1": [1],
            "GRAD2": [2],
            "plannar2": [2],
            "ALL": [0, 1, 2],
        }
        return sensor_mapping.get(sensortype)

    def _assert_sizes(self, train_size, valid_size, test_size=None):
        """Asserts that the sum of the data ratios is equal to 1."""
        if test_size is None:
            test_size = 0
        assert (
            sum((train_size, valid_size, test_size)) == 1
        ), "sum of data ratios must be equal to 1"

    def _leave_subjects_out_split(self, sizes, generator):
        """Leaves subjects out split."""
        indexes = [[], [], []]
        for i, split in enumerate(random_split(np.arange(self.n_subjects), sizes, generator)):
            indexes[i] = [
                idx for sub in split for idx in np.where(self.groups == sub)[0].tolist()
            ]
        return tuple(indexes)

    def _within_subject_split(self, sizes, generator):
        """Splits data within each subject."""
        indexes = []
        index_groups = [[] for _ in range(self.n_subjects)]
        for index, group in enumerate(self.groups):
            index_groups[group].append(index)
        for group in index_groups:
            indexes.append(random_split(group, sizes, generator))
        indexes = zip(*[random_split(group, sizes, generator) for group in index_groups])
        return tuple(sum(map(list, index), []) for index in indexes)

    def split_data(
        self, train_size: float = None, valid_size: float = None, test_size: float = None
    ):
        """
        Splits data into training, validation, and test sets.

        Parameters
        ----------
        train_size : float
            Train set size (%).
        valid_size : float
            Validation set size (%).
        test_size : float, optional
            Test set size (%). Defaults to None.

        Returns
        -------
        tuple
            Indices for the splits.
        """
        if train_size is None:
            train_size = self.split_sizes[0]
        if valid_size is None:
            valid_size = self.split_sizes[1]
            # if test_size is None:
            test_size = self.split_sizes[2]

        self._assert_sizes(train_size, valid_size, test_size)
        generator = torch.Generator().manual_seed(self.random_state)

        sizes = (train_size, valid_size, test_size)
        if self.lso:
            return self._leave_subjects_out_split(sizes, generator)
        elif self.groups is not None:
            return self._within_subject_split(sizes, generator)
        else:
            return random_split(np.arange(len(self)), sizes, generator)

    def torchDataset(self, index):
        """Returns a Torch dataset instance of the torch Dataset class for the given index."""
        return torch.utils.data.TensorDataset(self.data[index], self.targets[index])

    def set_targets(self, targets, target_labels=None):
        if type(targets[0]) in (str, np.str_):
            targets, target_labels = _string_to_int(targets, target_labels)

        self.targets = torch.Tensor(targets)
        self.target_labels = target_labels


class ContinuousDataset(EpochedDataset):
    """
    Creates a dataset from continuous data by loading continuous data and splitting it in segments according to set parameters.

    Parameters
    ----------
    window : float, optional
        Window size in seconds. Defaults to 2.
    overlap : float, optional
        Overlap between windows (0-1). Defaults to 0.
    offset : int, optional
        Offset in seconds. Defaults to 10.
    sfreq : float, optional
        Sampling frequency. Defaults to 500.
    n_subjects : int, optional
        Number of subjects. Defaults to None (all subjects).
    scaling : str, optional
        The scaling method. available options are "zscore" and "minmax". By default, minmax.
    n_samples : int, optional
        Number of samples per subject. Defaults to None.
    split_sizes(tuple or int), optional
        A tuple of (train_size, valid_size, test_size) for splits or a float <= 1,
        in which case the valid sizes and test sizes are deduced to be as half or the remaining.
    sensortype : str, optional
        Sensor type. Defaults to None.
    lso : bool, optional
        Leave subjects out. Defaults to False.
    random_state : int, optional
        Random state for reproducibility. Defaults to 0.

    Attributes
    ----------
    window : int
        Window size in seconds.
    overlap : float
        Overlap between windows.
    offset : int
        Offset in seconds.
    sfreq : int
        Sampling frequency.
    n_subjects : int
        Number of subjects.
    n_samples : int
        Number of samples per subject.
    data : torch.Tensor
        Data.
    targets : torch.Tensor
        targets.
    groups : list
        Groups.
    subject_list : list
        Subject list.

    Methods
    -------
    _load_sub(filepath)
        Loads a subject's data with windowing and overlap.
    """

    def __init__(
        self,
        window: int = 2,
        overlap: float = 0,
        offset: int = 10,
        sfreq: int = 500,
        n_subjects: int = None,
        scaling: str = "minmax",
        n_samples: int = None,
        split_sizes: tuple = (0.8, 0.1, 0.1),
        sensortype: str = None,
        lso: bool = False,
        random_state: int = 0,
    ) -> None:
        """
        Initializes the ContinuousDataset.

        Args:
        window (int): Window size in seconds.
        overlap (float): Overlap between windows.
        offset (int): Offset in seconds.
        sfreq (int): Sampling frequency.
        n_subjects (int): Number of subjects.
        scaling (str): The scaling method used.
        n_samples (int): Number of samples per subject.
        split_sizes(tuple or int): a tuple of (train_size, valid_size, test_size)
            for splits or a float <= 1, in which case the valid sizes and test sizes
            are deduced to be as half or the remaining.
        sensortype (str): Sensor type.
        lso (bool): Leave subjects out.
        random_state (int): Random state for reproducibility.
        """

        super().__init__(
            sfreq, n_subjects, scaling, n_samples, split_sizes, sensortype, lso, random_state
        )

        if scaling == "zscore":
            self.scaler = lambda x: zscore(x, axis=-1)
        elif scaling == "minmax":
            self.scaler = lambda x: (x - x.min()) / (x.max() - x.min())
        else:
            raise ValueError(f"{scaling} is an invalid scaling option.")
        assert 0 <= overlap < 1, "Overlap must be between 0 and 1."
        self.window = window
        self.overlap = overlap
        self.offset = offset

    def _load_sub(self, filepath: str) -> torch.Tensor:
        """
        Loads a subject's data with windowing and overlap.

        Parameters
        ----------
        filepath : str
            Path to the subject's data file.

        Returns
        -------
        torch.Tensor
            Loaded data.
        """
        try:
            step = int(self.window * self.sfreq * (1 - self.overlap))
            start = int(self.offset * self.sfreq)
            sub_data = np.load(filepath)
            if len(sub_data.shape) < 3:
                sub_data = sub_data[np.newaxis, :]
            data = []
            for i in range(start, sub_data.shape[-1], step):
                trial = sub_data[:, :, i : i + step]
                if trial.shape[-1] == step:
                    if not np.isnan(trial).any():
                        trial = [self.scaler(sensor_data) for sensor_data in trial]
                        data.append(trial)
        except IOError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        except ValueError:
            LOG.warning(f"There was a problem loading subject {filepath}")
            return None
        return torch.Tensor(np.array(data))


# For backwards compatibility purposes
RestDataset = ContinuousDataset
Dataset = EpochedDataset



import torch
import torch.nn as nn
import torch.nn.functional as F


class Conv2dWithConstraint(nn.Conv2d):
    def __init__(self, *args, max_norm=1, **kwargs):
        self.max_norm = max_norm
        super(Conv2dWithConstraint, self).__init__(*args, **kwargs)

    def forward(self, x):
        with torch.no_grad():
            self.weight.data = torch.renorm(
                self.weight.data, p=2, dim=0, maxnorm=self.max_norm
            )
        return super(Conv2dWithConstraint, self).forward(x)


class DeMixing(nn.Module):
    """
    Spatial demixing Layer.

    Parameters
    ----------
    size : int
        Number of output features.
    nonlin : callable
        Activation function.
    axis : int
        Axis to apply demixing.
    specs : dict
        Dictionary of layer specifications.
    """

    def __init__(
        self,
        size,
        nonlin=nn.Identity(),
        axis=-1,
        regularize="l1_lambda",
        bias_trainable=True,
        bias_const=0,
        unitnorm=False,
    ):
        super(DeMixing, self).__init__()
        self.size = size
        self.nonlin = nonlin
        self.axis = axis
        self.bias_trainable = bias_trainable
        self.bias_const = bias_const

        self.w = nn.Parameter(torch.empty((size,)), requires_grad=True)
        self.b_in = nn.Parameter(torch.empty((size,)), requires_grad=self.bias_trainable)

        nn.init.kaiming_uniform_(self.w, nonlinearity="relu")
        nn.init.constant_(self.b_in, self.bias_const)

        self.constraint = nn.functional.normalize if unitnorm else None

        self.reg = None
        if regularize == "l1_lambda":
            self.reg = nn.L1Loss()
        elif regularize == "l2_scope":
            self.reg = nn.MSELoss()

    def forward(self, x):
        """
        Applies demixing to input tensor.

        Args:
        x (torch.Tensor): Input tensor.

        Returns:
        torch.Tensor: Demixed tensor.
        """
        demix = torch.tensordot(x, self.w, dims=([self.axis], [0]))
        demix = self.nonlin(demix + self.b_in)
        if self.constraint:
            demix = self.constraint(demix, p=2, dim=1)
        return demix


class Flatten(nn.Module):
    """
    Flatten layer to connect feature extraction and classification parts of a network.

    Parameters
    ----------
    None

    Attributes
    ----------
    None

    Methods
    -------
    forward(x)
        Flattens the input tensor.
    """

    def forward(self, x):
        """Flattens the input tensor."""
        return x.view(x.size(0), -1)


class DepthwiseConv2d(nn.Module):
    """
    Depthwise separable convolutional layer.

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    kernel_size : int or tuple
        Size of the convolutional kernel.
    depthwise_multiplier : int, optional
        Multiplier for depthwise convolution. Defaults to 1.

    Attributes
    ----------
    depthwise : nn.Conv2d
        Depthwise convolutional layer.

    Methods
    -------
    forward(x)
        Applies depthwise convolution to the input tensor.
    """

    def __init__(self, in_channels, kernel_size, depthwise_multiplier=1, **kwargs):
        super(DepthwiseConv2d, self).__init__()
        self.depthwise = nn.Conv2d(
            in_channels,
            in_channels * depthwise_multiplier,
            kernel_size,
            groups=in_channels,
            **kwargs,
        )

    def forward(self, x):
        """Applies depthwise convolution to the input tensor."""
        return self.depthwise(x)


class SeparableConv2d(nn.Module):
    """
    Separable convolutional layer (depthwise + pointwise convolution).

    Parameters
    ----------
    in_channels : int
        Number of input channels.
    out_channels : int
        Number of output channels.
    kernel_size : int or tuple
        Size of the convolutional kernel.

    Attributes
    ----------
    depthwise : DepthwiseConv2d
        Depthwise convolutional layer.
    pointwise : nn.Conv2d
        Pointwise convolutional layer.

    Methods
    -------
    forward(x)
        Applies separable convolution to the input tensor.
    """

    def __init__(self, in_channels, out_channels, kernel_size, **kwargs):
        super(SeparableConv2d, self).__init__()
        self.depthwise = DepthwiseConv2d(in_channels, kernel_size, **kwargs)
        self.pointwise = nn.Conv2d(in_channels, out_channels, kernel_size=1, **kwargs)

    def forward(self, x):
        """Applies separable convolution to the input tensor."""
        return self.pointwise(self.depthwise(x))



import configparser
import configargparse

parser = configargparse.ArgParser()


def save_config(args: dict, config_filepath: str = "config.ini"):
    """Saves a dictionnary to a given path.

    Args:
        args (dict): The argument dictionnary to be saved to the config filepath.
        config_filepath (str): The filepath to be used to save the config file.
    """
    with open(config_filepath, "w") as conf:
        config_object = configparser.ConfigParser()
        config_object.add_section("config")
        for key, value in args.items():
            config_object.set("config", key.replace("_", "-"), str(value))
        config_object.write(conf)
    return


parser.add(
    "-c",
    "--config",
    is_config_file=True,
    default="../default.ini",
    help="config file path",
)
parser.add(
    "--testsplit",
    type=int,
    default=-1,
    help="Will remove the 20% holdout set by default and usit for cross-val. Using 5-Fold instead of 4-Fold.",
)
parser.add(
    "--flat",
    action="store_true",
    help="will flatten sensors dimension",
)
parser.add(
    "--dataset",
    default="rest",
    choices=["rest", "passive", "smt","mixed"],
    help="The camcan MEG dataset to load.",
)
parser.add(
    "--lso",
    action="store_true",
    help="wether or not to use Leave Subject Out when splitting data",
)
parser.add(
    "--randomsearch",
    action="store_true",
    help="Launches one cross-val on a subset of data or full random search depending on testsplit parameter",
)
parser.add(
    "--fold",
    default=-1,
    type=int,
    help="will only do a specific fold if specified. must be between 0 and 3, or 0 and 4 if testsplit option is true",
)
parser.add(
    "--net-option",
    default="meegnet",
    choices=[
        "custom",
        "MEEGNet",
        "meegnet",
        "EEGNet",
        "eegnet",
        "vgg",
        "VGG",
        "vanPutNet",
        "mlp",
        "MLP",
    ],
)
parser.add(
    "--epoched",
    action="store_true",
    help="Flag data as epoched if it is already epoched in the data files. otherwise the dataloader will segment data according to set parameters",
)
parser.add(
    "--crossval",
    action="store_true",
    help="wether to do a 4-FOLD cross-validation on the train+valid set.",
)
parser.add(
    "--n-samples",
    type=int,
    default=-1,
    help="limit of number of samples per subjects",
)
parser.add("-f", "--filters", default=8, type=int, help="The size of the first convolution")
parser.add(
    "--segment-length",
    type=float,
    default=2,
    help="The length (in seconds) of the segment to consider for resting-state data.",
)
parser.add(
    "--sfreq",
    type=int,
    default=500,
    help="The sampling frequency of the data, mainly for loading the correct data",
)
parser.add(
    "--maxpool",
    type=int,
    default=0,
    help="adds a maxpool layer in between conv layers",
)
parser.add(
    "--batchnorm",
    action="store_true",
    help="adds a batchnorm layer in between conv layers",
)
parser.add(
    "--permute-labels",
    action="store_true",
    help="Permutes the labesl in order to test for chance level",
)
parser.add(
    "--printmem",
    action="store_true",
    help="Shows RAM information before and during the data loading process.",
)
parser.add(
    "--log",
    action="store_true",
    help="stores all prints in a logfile instead of printing them",
)
parser.add(
    "--lr",
    type=float,
    default=0.00001,
    help="the starting learning rate of the optimizer",
)
parser.add(
    "--hlayers",
    type=int,
    default=1,
    help="number of hidden layers",
)
parser.add(
    "--patience",
    type=int,
    default=20,
    help="patience for early stopping",
)
parser.add(
    "--model-name",
    type=str,
    default="net",
    help="Name of the network for file_save",
)
parser.add(
    "--num-workers",
    type=int,
    default=4,
    help="number of workers to load data while gpu is processing",
)
parser.add(
    "--train-size",
    type=float,
    default=0.8,
    help="The proportion of data to use in the train set",
)
parser.add(
    "--model-path",
    type=str,
    default=None,
    help="The default path to save all computed data, model and visualisations.",
)
parser.add(
    "--save-path",
    type=str,
    default=".",
    help="The default path to save all computed data, model and visualisations.",
)
parser.add(
    "--raw-path",
    type=str,
    default=None,
    help="The path where the raw data can be found.",
)
parser.add(
    "--visu-path",
    type=str,
    default=None,
    help="The path where the visualisation matrices wil be saved to and loaded from.",
)
parser.add(
    "--processed-path",
    type=str,
    default=None,
    help="The path where the data samples can be found.",
)
parser.add(
    "--seed",
    default=42,
    type=int,
    help="Seed to use for random splits.",
)
parser.add(
    "--max-subj",
    default=1000,
    type=int,
    help="maximum number of subjects to use (1000 uses all subjects)",
)
parser.add(
    "-s",
    "--sensors",
    default="MAG",
    choices=["GRAD", "MAG", "ALL"],
    help="The type of sensors to keep, default=MAG",
)
parser.add(
    "--feature",
    default="temporal",
    choices=["temporal", "bands", "bins"],
    help="Data type to use.",
)
parser.add(
    "-b",
    "--batch-size",
    default=128,
    type=int,
    help="The batch size used for learning.",
)
parser.add(
    "-d",
    "--dropout",
    default=0.25,
    type=float,
    help="The dropout rate of the linear layers",
)
parser.add(
    "--times",
    action="store_true",
    help="Instead of running the training etc, run a series of test in order to choose best set of workers and batch sizes to get faster epochs.",
)
parser.add(
    "--chunkload",
    action="store_true",
    help="Chunks the data and loads data batch per batch. Will be slower but is necessary when RAM size is too low to handle whole dataset.",
)
parser.add(
    "--debug",
    action="store_true",
    help="loads dummy data in the net to ensure everything is working fine",
)
parser.add(
    "--dropout-option",
    default="same",
    choices=["same", "double", "inverted"],
    help="sets if the first dropout and the second are the same or if the first one or the second one should be bigger",
)
parser.add("-l", "--linear", default=100, type=int, help="The size of the second linear layer")
parser.add(
    "-n",
    "--nchan",
    default=100,
    type=int,
    help="the number of channels for the first convolution, the other channel numbers scale with this one.",
)
parser.add(
    "--compute-psd",
    action="store_true",
    help="wether or not to to compute psd using saliency windows in the compute_saliency_maps.py script.",
)
parser.add(
    "--w-size",
    type=int,
    default=300,
    help="The window size for saliency based psd computation in the compute_saliency_maps.py script.",
)
parser.add(
    "--overlap",
    type=float,
    default=0,
    help="the overlap value between segments for continous data.",
)
parser.add(
    "--confidence",
    type=float,
    default=0.98,
    help="the confidence threshold needed for a trial to be selected for visualisation in the compute_ and visu_saliency_maps.py script.",
)



import os
import mne
import logging
import copy
import torch
from mne.viz import plot_topomap
from torch.autograd import Variable
from torchvision import models
import numpy as np
from PIL import Image
import matplotlib as mpl
import matplotlib.pyplot as plt
import matplotlib.pyplot as plt
import matplotlib.cm as mpl_color_map
from matplotlib.gridspec import GridSpec
from collections.abc import Iterable
from joblib import Parallel, delayed
from meegnet.utils import compute_psd, cuda_check
from pytorch_grad_cam import GuidedBackpropReLUModel
from pytorch_grad_cam import GradCAM
from pytorch_grad_cam.utils.model_targets import ClassifierOutputTarget
from pytorch_grad_cam.utils.image import show_cam_on_image, deprocess_image, preprocess_image
import torch.nn.functional as F
import cv2
from mpl_toolkits.axes_grid1 import make_axes_locatable


LOG = logging.getLogger("meegnet")
mne.set_log_level(False)


def compute_saliency_maps(
    dataset,
    net,
    sal_path,
    threshold=0.95,
    labels=None,
    epoched=False,
):

    device = cuda_check()
    GBP = GuidedBackpropReLUModel(net, device=device)

    if labels is None:
        labels = dataset.target_labels

    # Load all trials and corresponding labels for a specific subject.
    data = dataset.data
    targets = dataset.targets
    if epoched:
        target_saliencies = [[[], []], [[], []]]
    else:
        target_saliencies = [[], []]

    # For each of those trial with associated label:
    for trial, label in zip(data, targets):
        X = trial
        while len(X.shape) < 4:
            X = X[np.newaxis, :]
        X = X.to(device)
        # Compute predictions of the trained network, and confidence
        preds = torch.nn.Softmax(dim=1)(net(X)).detach().cpu()
        pred = preds.argmax().item()
        confidence = preds.max()
        label = int(label)

        # If the confidence reaches desired treshhold (given by args.confidence)
        if confidence >= threshold and pred == label:
            # Compute Guided Back-propagation for given label projected on given data X
            guided_grads = GBP(X.to(device), label)
            guided_grads = np.rollaxis(guided_grads, 2, 0)
            # Compute saliencies
            pos_saliency, neg_saliency = get_positive_negative_saliency(guided_grads)

            # Depending on the task, add saliencies in lists
            if epoched:
                target_saliencies[label][0].append(pos_saliency)
                target_saliencies[label][1].append(neg_saliency)
            else:
                target_saliencies[0].append(pos_saliency)
                target_saliencies[1].append(neg_saliency)
    # With all saliencies computed, we save them in the specified save-path
    n_saliencies = 0
    n_saliencies += sum([len(e) for e in target_saliencies[0]])
    n_saliencies += sum([len(e) for e in target_saliencies[1]])

    sub = dataset.subject_list[0]
    LOG.info(f"{n_saliencies} saliency maps computed for {sub}")
    for j, sal_type in enumerate(("pos", "neg")):
        if epoched:
            for i, label in enumerate(labels):
                sal_filepath = os.path.join(
                    sal_path,
                    f"{sub}_{labels[i]}_{sal_type}_sal_{threshold}confidence.npy",
                )
                np.save(sal_filepath, np.array(target_saliencies[i][j]))
        else:
            lab = "" if not epoched else f"_{labels[label]}"
            sal_filepath = os.path.join(
                sal_path,
                f"{sub}{lab}_{sal_type}_sal_{threshold}confidence.npy",
            )
            np.save(sal_filepath, np.array(target_saliencies[j]))


def convert_to_grayscale(im_as_arr):
    """
        Converts 3d image to grayscale

    Args:
        im_as_arr (numpy arr): RGB image with shape (D,W,H)

    returns:
        grayscale_im (numpy_arr): Grayscale image with shape (1,W,D)
    """
    grayscale_im = np.sum(np.abs(im_as_arr), axis=0)
    im_max = np.percentile(grayscale_im, 99)
    im_min = np.min(grayscale_im)
    grayscale_im = np.clip((grayscale_im - im_min) / (im_max - im_min), 0, 1)
    grayscale_im = np.expand_dims(grayscale_im, axis=0)
    return grayscale_im


def save_gradient_images(gradient, file_name):
    """
        Exports the original gradient image

    Args:
        gradient (np arr): Numpy array of the gradient with shape (3, 224, 224)
        file_name (str): File name to be exported
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Normalize
    gradient = gradient - gradient.min()
    gradient /= gradient.max()
    # Save image
    # path_to_file = os.path.join("../results", file_name + ".jpg")
    path_to_file = os.path.join(file_name + ".jpg")
    save_image(gradient, path_to_file)


def save_class_activation_images(org_img, activation_map, file_name):
    """
        Saves cam activation map and activation map on the original image

    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        file_name (str): File name of the exported image
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Grayscale activation map
    heatmap, heatmap_on_image = apply_colormap_on_image(org_img, activation_map, "hsv")
    # Save colored heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Heatmap.png")
    save_image(heatmap, path_to_file)
    # Save heatmap on iamge
    path_to_file = os.path.join("../results", file_name + "_Cam_On_Image.png")
    save_image(heatmap_on_image, path_to_file)
    # SAve grayscale heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Grayscale.png")
    save_image(activation_map, path_to_file)


def apply_colormap_on_image(org_im, activation, colormap_name):
    """
        Apply heatmap on image
    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        colormap_name (str): Name of the colormap
    """
    # Get colormap
    color_map = mpl_color_map.get_cmap(colormap_name)
    no_trans_heatmap = color_map(activation)
    # Change alpha channel in colormap to make sure original image is displayed
    heatmap = copy.copy(no_trans_heatmap)
    heatmap[:, :, 3] = 0.4
    heatmap = Image.fromarray((heatmap * 255).astype(np.uint8))
    no_trans_heatmap = Image.fromarray((no_trans_heatmap * 255).astype(np.uint8))

    # Apply heatmap on iamge
    heatmap_on_image = Image.new("RGBA", org_im.size)
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, org_im.convert("RGBA"))
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, heatmap)
    return no_trans_heatmap, heatmap_on_image


def format_np_output(np_arr):
    """
        This is a (kind of) bandaid fix to streamline saving procedure.
        It converts all the outputs to the same format which is 3xWxH
        with using sucecssive if clauses.
    Args:
        im_as_arr (Numpy array): Matrix of shape 1xWxH or WxH or 3xWxH
    """
    # Phase/Case 1: The np arr only has 2 dimensions
    # Result: Add a dimension at the beginning
    if len(np_arr.shape) == 2:
        np_arr = np.expand_dims(np_arr, axis=0)
    # Phase/Case 2: Np arr has only 1 channel (assuming first dim is channel)
    # Result: Repeat first channel and convert 1xWxH to 3xWxH
    if np_arr.shape[0] == 1:
        np_arr = np.repeat(np_arr, 3, axis=0)
    # Phase/Case 3: Np arr is of shape 3xWxH
    # Result: Convert it to WxHx3 in order to make it saveable by PIL
    if np_arr.shape[0] == 3:
        np_arr = np_arr.transpose(1, 2, 0)
    # Phase/Case 4: NP arr is normalized between 0-1
    # Result: Multiply with 255 and change type to make it saveable by PIL
    if np.max(np_arr) <= 1:
        np_arr = (np_arr * 255).astype(np.uint8)
    return np_arr


def save_image(im, path):
    """
        Saves a numpy matrix or PIL image as an image
    Args:
        im_as_arr (Numpy array): Matrix of shape DxWxH
        path (str): Path to the image
    """
    if isinstance(im, (np.ndarray, np.generic)):
        im = format_np_output(im)
        im = Image.fromarray(im)
    im.save(path)


def preprocess_image(pil_im, resize_im=True):
    """
        Processes image for CNNs

    Args:
        PIL_img (PIL_img): PIL Image or numpy array to process
        resize_im (bool): Resize to 224 or not
    returns:
        im_as_var (torch variable): Variable that contains processed float tensor
    """
    # mean and std list for channels (Imagenet)
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # ensure or transform incoming image to PIL image
    if type(pil_im) != Image.Image:
        try:
            pil_im = Image.fromarray(pil_im)
        except Exception as e:
            print("could not transform PIL_img to a PIL Image object. Please check input.")

    # Resize image
    if resize_im:
        pil_im = pil_im.resize((224, 224), Image.ANTIALIAS)

    im_as_arr = np.float32(pil_im)
    im_as_arr = im_as_arr.transpose(2, 0, 1)  # Convert array to D,W,H
    # Normalize the channels
    for channel, _ in enumerate(im_as_arr):
        im_as_arr[channel] /= 255
        im_as_arr[channel] -= mean[channel]
        im_as_arr[channel] /= std[channel]
    # Convert to float tensor
    im_as_ten = torch.from_numpy(im_as_arr).float()
    # Add one more channel to the beginning. Tensor shape = 1,3,224,224
    im_as_ten.unsqueeze_(0)
    # Convert to Pytorch variable
    im_as_var = Variable(im_as_ten, requires_grad=True)
    return im_as_var


def recreate_image(im_as_var):
    """
        Recreates images from a torch variable, sort of reverse preprocessing
    Args:
        im_as_var (torch variable): Image to recreate
    returns:
        recreated_im (numpy arr): Recreated image in array
    """
    reverse_mean = [-0.485, -0.456, -0.406]
    reverse_std = [1 / 0.229, 1 / 0.224, 1 / 0.225]
    recreated_im = copy.copy(im_as_var.data.numpy()[0])
    for c in range(3):
        recreated_im[c] /= reverse_std[c]
        recreated_im[c] -= reverse_mean[c]
    recreated_im[recreated_im > 1] = 1
    recreated_im[recreated_im < 0] = 0
    recreated_im = np.round(recreated_im * 255)

    recreated_im = np.uint8(recreated_im).transpose(1, 2, 0)
    return recreated_im


def choose_best_window(data, fs=500, w_size=300):
    """
    data: array
        Must be of size k x n_samples. k can be sensor dimension or trial dimension.
    w_size: int
        The size of the window in ms
    """
    # Generate masks over time samples where sample is superior to 2x std of the vector
    masks = [dat >= (np.mean(dat) + np.std(dat) * 2) for dat in data]
    w_size = int(w_size * fs / 1000)
    best_window_idx = Parallel(n_jobs=-1)(delayed(best_window)(mask, w_size) for mask in masks)
    return best_window_idx


def best_window(mask, w_size):
    # create an array of all the possible windows in the trial with their mask values
    windows = np.array([mask[i : i + w_size] for i in range(len(mask) - w_size)])
    # get the sum of mask values in each windows. High number indicates lots of True values
    values = [sum(window) for window in windows]
    if max(values) == 0:
        return None
    # best window is where the values are the highest
    best_window_index = np.where(values == max(values))[0]
    # below is to handle if there are multiple best windows with the same values, in which case we
    # try to check if they overlap, and keep only if they are separated.
    if len(best_window_index) > 1:
        idx_range = best_window_index[-1] - best_window_index[0]
        # duration betweeen first and last window must be w_size/2
        if idx_range <= int(w_size / 2):
            # Then best would be in the middle of all this
            best = int(len(best_window_index) / 2)
            return best_window_index[best]
        elif idx_range <= w_size:
            return (best_window_index[0], best_window_index[-1])
        else:
            best = int(len(best_window_index) / 2)
            return (best_window_index[0], best, best_window_index[-1])
    else:
        # only one window, just add it.
        return best_window_index[0]


def compute_single_sal_psd(index, trial, w_size, fs):
    if isinstance(index, Iterable):
        tmp = []
        for idx in index:
            window = trial[idx : idx + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            tmp.append(compute_psd(window, fs=fs))
        return np.mean(tmp, axis=0)
    else:
        if index is not None:
            window = trial[index : index + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            return compute_psd(window, fs=fs)
        else:
            return [None] * 7


def compute_saliency_based_psd(saliency, trial, w_size, fs):
    chan_data = []
    for i, chan in enumerate(saliency):
        windows_idx = choose_best_window(chan, fs, w_size)
        transformed_data = Parallel(n_jobs=1)(
            delayed(compute_single_sal_psd)(index, trial[i, j], w_size, fs)
            for j, index in enumerate(windows_idx)
        )
        chan_data.append(np.array(transformed_data).squeeze())
    return np.array(chan_data)


def get_positive_negative_saliency(gradient):
    """
        Generates positive and negative saliency maps based on the gradient
    Args:
        gradient (numpy arr): Gradient of the operation to visualize

    returns:
        pos_saliency ( )
    """
    pos_saliency = np.maximum(0, gradient) / gradient.max()
    neg_saliency = np.maximum(0, -gradient) / -gradient.min()
    return pos_saliency, neg_saliency


def get_example_params(example_index):
    """
        Gets used variables for almost all visualizations, like the image, model etc.

    Args:
        example_index (int): Image id to use from examples

    returns:
        original_image (numpy arr): Original image read from the file
        prep_img (numpy_arr): Processed image
        target_class (int): Target class for the image
        file_name_to_export (string): File name to export the visualizations
        pretrained_model(Pytorch model): Model to use for the operations
    """
    # Pick one of the examples
    example_list = (
        ("../input_images/snake.jpg", 56),
        ("../input_images/cat_dog.png", 243),
        ("../input_images/spider.png", 72),
    )
    img_path = example_list[example_index][0]
    target_class = example_list[example_index][1]
    file_name_to_export = img_path[img_path.rfind("/") + 1 : img_path.rfind(".")]
    # Read image
    original_image = Image.open(img_path).convert("RGB")
    # Process image
    prep_img = preprocess_image(original_image)
    # Define model
    pretrained_model = models.alexnet(pretrained=True)
    return (
        original_image,
        prep_img,
        target_class,
        file_name_to_export,
        pretrained_model,
    )


def convert_to_grayscale(im_as_arr):
    """
        Converts 3d image to grayscale

    Args:
        im_as_arr (numpy arr): RGB image with shape (D,W,H)

    returns:
        grayscale_im (numpy_arr): Grayscale image with shape (1,W,D)
    """
    grayscale_im = np.sum(np.abs(im_as_arr), axis=0)
    im_max = np.percentile(grayscale_im, 99)
    im_min = np.min(grayscale_im)
    grayscale_im = np.clip((grayscale_im - im_min) / (im_max - im_min), 0, 1)
    grayscale_im = np.expand_dims(grayscale_im, axis=0)
    return grayscale_im


def save_gradient_images(gradient, file_name):
    """
        Exports the original gradient image

    Args:
        gradient (np arr): Numpy array of the gradient with shape (3, 224, 224)
        file_name (str): File name to be exported
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Normalize
    gradient = gradient - gradient.min()
    gradient /= gradient.max()
    # Save image
    # path_to_file = os.path.join("../results", file_name + ".jpg")
    path_to_file = os.path.join(file_name + ".jpg")
    save_image(gradient, path_to_file)


def save_class_activation_images(org_img, activation_map, file_name):
    """
        Saves cam activation map and activation map on the original image

    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        file_name (str): File name of the exported image
    """
    if not os.path.exists("../results"):
        os.makedirs("../results")
    # Grayscale activation map
    heatmap, heatmap_on_image = apply_colormap_on_image(org_img, activation_map, "hsv")
    # Save colored heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Heatmap.png")
    save_image(heatmap, path_to_file)
    # Save heatmap on iamge
    path_to_file = os.path.join("../results", file_name + "_Cam_On_Image.png")
    save_image(heatmap_on_image, path_to_file)
    # SAve grayscale heatmap
    path_to_file = os.path.join("../results", file_name + "_Cam_Grayscale.png")
    save_image(activation_map, path_to_file)


def apply_colormap_on_image(org_im, activation, colormap_name):
    """
        Apply heatmap on image
    Args:
        org_img (PIL img): Original image
        activation_map (numpy arr): Activation map (grayscale) 0-255
        colormap_name (str): Name of the colormap
    """
    # Get colormap
    color_map = mpl_color_map.get_cmap(colormap_name)
    no_trans_heatmap = color_map(activation)
    # Change alpha channel in colormap to make sure original image is displayed
    heatmap = copy.copy(no_trans_heatmap)
    heatmap[:, :, 3] = 0.4
    heatmap = Image.fromarray((heatmap * 255).astype(np.uint8))
    no_trans_heatmap = Image.fromarray((no_trans_heatmap * 255).astype(np.uint8))

    # Apply heatmap on iamge
    heatmap_on_image = Image.new("RGBA", org_im.size)
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, org_im.convert("RGBA"))
    heatmap_on_image = Image.alpha_composite(heatmap_on_image, heatmap)
    return no_trans_heatmap, heatmap_on_image


def format_np_output(np_arr):
    """
        This is a (kind of) bandaid fix to streamline saving procedure.
        It converts all the outputs to the same format which is 3xWxH
        with using sucecssive if clauses.
    Args:
        im_as_arr (Numpy array): Matrix of shape 1xWxH or WxH or 3xWxH
    """
    # Phase/Case 1: The np arr only has 2 dimensions
    # Result: Add a dimension at the beginning
    if len(np_arr.shape) == 2:
        np_arr = np.expand_dims(np_arr, axis=0)
    # Phase/Case 2: Np arr has only 1 channel (assuming first dim is channel)
    # Result: Repeat first channel and convert 1xWxH to 3xWxH
    if np_arr.shape[0] == 1:
        np_arr = np.repeat(np_arr, 3, axis=0)
    # Phase/Case 3: Np arr is of shape 3xWxH
    # Result: Convert it to WxHx3 in order to make it saveable by PIL
    if np_arr.shape[0] == 3:
        np_arr = np_arr.transpose(1, 2, 0)
    # Phase/Case 4: NP arr is normalized between 0-1
    # Result: Multiply with 255 and change type to make it saveable by PIL
    if np.max(np_arr) <= 1:
        np_arr = (np_arr * 255).astype(np.uint8)
    return np_arr


def save_image(im, path):
    """
        Saves a numpy matrix or PIL image as an image
    Args:
        im_as_arr (Numpy array): Matrix of shape DxWxH
        path (str): Path to the image
    """
    if isinstance(im, (np.ndarray, np.generic)):
        im = format_np_output(im)
        im = Image.fromarray(im)
    im.save(path)


def preprocess_image(pil_im, resize_im=True):
    """
        Processes image for CNNs

    Args:
        PIL_img (PIL_img): PIL Image or numpy array to process
        resize_im (bool): Resize to 224 or not
    returns:
        im_as_var (torch variable): Variable that contains processed float tensor
    """
    # mean and std list for channels (Imagenet)
    mean = [0.485, 0.456, 0.406]
    std = [0.229, 0.224, 0.225]

    # ensure or transform incoming image to PIL image
    if type(pil_im) != Image.Image:
        try:
            pil_im = Image.fromarray(pil_im)
        except Exception as e:
            print("could not transform PIL_img to a PIL Image object. Please check input.")

    # Resize image
    if resize_im:
        pil_im = pil_im.resize((224, 224), Image.ANTIALIAS)

    im_as_arr = np.float32(pil_im)
    im_as_arr = im_as_arr.transpose(2, 0, 1)  # Convert array to D,W,H
    # Normalize the channels
    for channel, _ in enumerate(im_as_arr):
        im_as_arr[channel] /= 255
        im_as_arr[channel] -= mean[channel]
        im_as_arr[channel] /= std[channel]
    # Convert to float tensor
    im_as_ten = torch.from_numpy(im_as_arr).float()
    # Add one more channel to the beginning. Tensor shape = 1,3,224,224
    im_as_ten.unsqueeze_(0)
    # Convert to Pytorch variable
    im_as_var = Variable(im_as_ten, requires_grad=True)
    return im_as_var


def recreate_image(im_as_var):
    """
        Recreates images from a torch variable, sort of reverse preprocessing
    Args:
        im_as_var (torch variable): Image to recreate
    returns:
        recreated_im (numpy arr): Recreated image in array
    """
    reverse_mean = [-0.485, -0.456, -0.406]
    reverse_std = [1 / 0.229, 1 / 0.224, 1 / 0.225]
    recreated_im = copy.copy(im_as_var.data.numpy()[0])
    for c in range(3):
        recreated_im[c] /= reverse_std[c]
        recreated_im[c] -= reverse_mean[c]
    recreated_im[recreated_im > 1] = 1
    recreated_im[recreated_im < 0] = 0
    recreated_im = np.round(recreated_im * 255)

    recreated_im = np.uint8(recreated_im).transpose(1, 2, 0)
    return recreated_im


def choose_best_window(data, fs=500, w_size=300):
    """
    data: array
        Must be of size k x n_samples. k can be sensor dimension or trial dimension.
    w_size: int
        The size of the window in ms
    """
    # Generate masks over time samples where sample is superior to 2x std of the vector
    masks = [dat >= (np.mean(dat) + np.std(dat) * 2) for dat in data]
    w_size = int(w_size * fs / 1000)
    best_window_idx = Parallel(n_jobs=-1)(delayed(best_window)(mask, w_size) for mask in masks)
    return best_window_idx


def best_window(mask, w_size):
    # create an array of all the possible windows in the trial with their mask values
    windows = np.array([mask[i : i + w_size] for i in range(len(mask) - w_size)])
    # get the sum of mask values in each windows. High number indicates lots of True values
    values = [sum(window) for window in windows]
    if max(values) == 0:
        return None
    # best window is where the values are the highest
    best_window_index = np.where(values == max(values))[0]
    # below is to handle if there are multiple best windows with the same values, in which case we
    # try to check if they overlap, and keep only if they are separated.
    if len(best_window_index) > 1:
        idx_range = best_window_index[-1] - best_window_index[0]
        # duration betweeen first and last window must be w_size/2
        if idx_range <= int(w_size / 2):
            # Then best would be in the middle of all this
            best = int(len(best_window_index) / 2)
            return best_window_index[best]
        elif idx_range <= w_size:
            return (best_window_index[0], best_window_index[-1])
        else:
            best = int(len(best_window_index) / 2)
            return (best_window_index[0], best, best_window_index[-1])
    else:
        # only one window, just add it.
        return best_window_index[0]


def compute_single_sal_psd(index, trial, w_size, fs):
    if isinstance(index, Iterable):
        tmp = []
        for idx in index:
            window = trial[idx : idx + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            tmp.append(compute_psd(window, fs=fs))
        return np.mean(tmp, axis=0)
    else:
        if index is not None:
            window = trial[index : index + w_size]
            while len(window.shape) < 3:
                window = window[np.newaxis, :]
            return compute_psd(window, fs=fs)
        else:
            return [None] * 7


def compute_saliency_based_psd(saliency, trial, w_size, fs):
    chan_data = []
    for i, chan in enumerate(saliency):
        windows_idx = choose_best_window(chan, fs, w_size)
        transformed_data = Parallel(n_jobs=1)(
            delayed(compute_single_sal_psd)(index, trial[i, j], w_size, fs)
            for j, index in enumerate(windows_idx)
        )
        chan_data.append(np.array(transformed_data).squeeze())
    return np.array(chan_data)


def get_positive_negative_saliency(gradient):
    """
        Generates positive and negative saliency maps based on the gradient
    Args:
        gradient (numpy arr): Gradient of the operation to visualize

    returns:
        pos_saliency ( )
    """
    pos_saliency = np.maximum(0, gradient) / gradient.max()
    neg_saliency = np.maximum(0, -gradient) / -gradient.min()
    return pos_saliency, neg_saliency


def get_example_params(example_index):
    """
        Gets used variables for almost all visualizations, like the image, model etc.

    Args:
        example_index (int): Image id to use from examples

    returns:
        original_image (numpy arr): Original image read from the file
        prep_img (numpy_arr): Processed image
        target_class (int): Target class for the image
        file_name_to_export (string): File name to export the visualizations
        pretrained_model(Pytorch model): Model to use for the operations
    """
    # Pick one of the examples
    example_list = (
        ("../input_images/snake.jpg", 56),
        ("../input_images/cat_dog.png", 243),
        ("../input_images/spider.png", 72),
    )
    img_path = example_list[example_index][0]
    target_class = example_list[example_index][1]
    file_name_to_export = img_path[img_path.rfind("/") + 1 : img_path.rfind(".")]
    # Read image
    original_image = Image.open(img_path).convert("RGB")
    # Process image
    prep_img = preprocess_image(original_image)
    # Define model
    pretrained_model = models.alexnet(pretrained=True)
    return (
        original_image,
        prep_img,
        target_class,
        file_name_to_export,
        pretrained_model,
    )


def make_gif(image_list, output=None, duration=100, loop=0):
    if output is None:
        output = ".".join(image_list[0].split(".")[:-1] + ["gif"])
    frames = [Image.open(image) for image in image_list]
    frame_one = frames[0]
    frame_one.save(
        output,
        format="GIF",
        append_images=frames,
        save_all=True,
        duration=duration,
        loop=loop,
    )


def avg_range(arr: list):
    length = int(len(arr) / 2)
    return rec_avg_range(list(arr), length)


def rec_avg_range(arr: list, length: int):
    if len(arr) <= 1:
        return 0
    argmax = np.argmax(arr)
    max = arr.pop(argmax)
    argmin = np.argmin(arr)
    min = arr.pop(argmin)
    range = (max - min) / length
    return range + rec_avg_range(arr, length)


def _unit_test_avg_range():
    liste = [18, -8, -6, 3, -10, 15, 0]
    assert avg_range(liste) == 20


def generate_saliency_figure(
    saliencies: dict,
    info,
    save_path: str = "",
    suffix: str = "",
    title: str = "",
    sensors: list = [""],
    sfreq=500,
    edge=100,
    cmap="coolwarm",
    stim_tick=None,
    show=False,
    outlines=None,
    topomap="window",
):
    """
    Generates a figure visualizing saliency maps for MEG data.

    This function creates a grid of images showing the saliency maps for
    types of stimuli (e.g., image and sound) and sensor channels (e.g., MAG, GRAD1, GRAD2).
    It also plots a topomap for the maximum saliency index along with a color bar.

    Parameters
    ----------
    saliencies : dict
        Dictionary containing saliency maps. Keys should correspond to different types
        of stimuli (e.g., "image", "sound"), and values should be numpy arrays of shape
        3 x sensors x samples, representing the saliency maps for each channel.
    save_path : str, optional
        Path to save the generated figure. Default is an empty string, which means the
        figure will not be saved automatically.
    suffix : str, optional
        Suffix to append to the filename when saving the figure. Default is an empty string.
    title : str, optional
        Title for the figure. Default is an empty string.
    sensors : list, optional
        List of sensor types to include in the visualization. Default is [""].
    sfreq : int, optional
        Sampling frequency for computation of  xticks. Default is  500.
    cmap : str, optional
        Colormap to use for displaying the topo- and saliency maps. Default is "coolwarm".
    stim_tick : int, optional
        Tick position for the stimulus event in the time axis. Default is None,
    show : bool, optional
        Wether to show the figure or not. Useful for ipynb.
    topomap : str, optional
        Must be "window", "average" or "timing". if " average, the average saliency across time is used. If
        timing, then the saliency at to the highest saliency timing is used. if "window" is used,
        then the saliency window around the max saliency timing is used.

    Returns
    -------
    None

    Notes
    -----
    The function assumes that the input saliency maps are normalized to have zero mean and
    unit variance. The colormap used for displaying the saliency maps is "coolwarm" by default,
    but can be changed with the `cmap` parameter.

    The function creates a grid layout with a subplot for each sensor channel and a subplot for the
    topomap. The grid layout is dynamically adjusted based on the number of sensors.

    The function does not handle exceptions that may occur during the plotting process, such as issues with
    file I/O or invalid input data.
    """

    assert topomap in (
        "timing",
        "average",
        "window",
    ), f"{topomap} is not a valid option for the topomap parameter. ('timing', 'average', 'window')"

    block_size = 3
    tick_ratio = 1000 / sfreq
    padding = int(edge / tick_ratio)
    if stim_tick is not None:
        assert stim_tick > edge, "Can't have stim_tick lower than the edge parameter."
    if suffix != "" and not suffix.endswith("_"):
        suffix += "_"
    n_blocs = len(sensors)  # number of blocs of figures in a line
    n_lines = len(saliencies)  # number of lines for the pyplot figure
    n_cols = n_blocs * block_size + 1  # number of columns for the pyplot figure
    grid = GridSpec(n_lines, n_cols)
    fig = plt.figure(figsize=(n_cols * 2, n_lines * 2))
    plt.title(title)
    plt.axis("off")
    axes = []
    # First pass to gather vlim values:
    for i, label in enumerate(saliencies.keys()):
        vlim = 0
        gradient = copy.copy(saliencies[label].squeeze())
        assert (
            len(gradient) == n_blocs
        ), "Can't generate figures for all sensors, check if the saliencies have been properly computed."
        gradient /= np.abs(gradient).max()
        for j, sensor_type in zip(range(0, n_blocs * block_size, block_size), sensors):
            idx = j // block_size
            length = gradient[idx].shape[1]
            grads = copy.copy(gradient[idx])[:, padding : length - padding]
            # segment_length = grads.shape[1]
            # mid_slice = (0, segment_length)
            # gradmeans = grads[:, mid_slice[0] : mid_slice[1]].mean(axis=1)[:, np.newaxis]
            # grads -= gradmeans  # We remove mean accross time to make the variations accross time pop-up more
            vmax = grads.max()
            vmin = grads.min()
            vlim_curr = max(abs(vmax), abs(vmin))
            if vlim_curr > vlim:
                vlim = vlim_curr

        for j, sensor_type in zip(range(0, n_blocs * block_size, block_size), sensors):
            idx = j // block_size
            # grads = gradient[idx]
            # In an attempt to remove the edge effect:
            # We remove the first and last edge points -> therefore tick is moved to 25 (was 75)
            length = gradient[idx].shape[1]
            grads = gradient[idx][:, padding : length - padding]

            segment_length = grads.shape[1]
            # We add the mid_slice variable in an attempt to tackle the edge effects by removing mean from center values for example
            # But it was uneffective. This could still be useful so we leave it here...
            # mid_slice = (0, segment_length)
            # mid_slice = (int(segment_length / 4), int(3 * segment_length / 4))
            # gradmeans = grads[:, mid_slice[0] : mid_slice[1]].mean(axis=1)[:, np.newaxis]
            # grads -= gradmeans  # We remove mean accross time to make the variations accross time pop-up more
            n_sensors = grads.shape[0]
            max_idx = np.unravel_index(abs(grads).argmax(), grads.shape)[1]
            # max_idx = np.argmax([avg_range(arr) for arr in grads.T])
            # max_idx = np.argmax(np.mean(grads, axis=0))
            axes.append(fig.add_subplot(grid[i, j : j + 2]))
            plt.imshow(
                grads,
                interpolation="nearest",
                aspect=1,
                vmin=-vlim,
                vmax=vlim,
                cmap=cmap,
            )
            axes[-1].spines["top"].set_visible(False)
            axes[-1].spines["right"].set_visible(False)
            axes[-1].yaxis.tick_right()

            x_ticks = [0, int(segment_length / 2), segment_length]

            if stim_tick is not None:
                stim_tick_index = int(stim_tick / tick_ratio)
                x_ticks.append(stim_tick_index)
                plt.axvline(x=stim_tick_index, color="black", linestyle="--", linewidth=1)
            else:
                stim_tick_index = 0

            if topomap != "average":
                plt.axvline(x=max_idx, color="green", linestyle="--", linewidth=1)
                x_ticks.append(max_idx)

            x_ticks = sorted(x_ticks)
            tick_labels = [(x_tick - stim_tick_index) * tick_ratio for x_tick in x_ticks]
            tick_labels[0] += edge
            tick_labels[-1] -= edge

            plt.xticks(x_ticks, tick_labels, fontsize=8)
            plt.yticks([0, n_sensors], [n_sensors, 0])

            if j == 0:
                axes[-1].text(-50, 50, label, ha="left", va="center", rotation="vertical")
            if idx == n_blocs - 1:
                axes[-1].yaxis.set_label_position("right")
                plt.ylabel("sensors")
            if i == 0:
                plt.title(sensor_type)
            if i == 1:
                plt.xlabel("time (ms)")
            axes.append(fig.add_subplot(grid[i, j + 2]))
            if topomap == "timing":
                data = grads[:, max_idx]
            elif topomap == "window":
                start = max_idx - int(segment_length / 8)
                end = max_idx + int(segment_length / 8)
                if start < 0:
                    start = 0
                    end = start + int(segment_length / 4)
                elif end > segment_length:
                    end = segment_length
                    start = block_size * int(segment_length / 4)
                data = grads[:, start:end].mean(axis=1)
            else:
                data = grads.mean(axis=1)

            if isinstance(info, str):
                linfo = np.load(
                    os.path.join(info, f"camcan_info_{sensor_type.lower()}.npy"),
                    allow_pickle=True,
                ).tolist()
            else:
                linfo = info
            im, _ = plot_topomap(
                data.ravel(),
                linfo,
                res=300,
                cmap=cmap,
                vlim=(-vlim, vlim),
                show=False,
                contours=0,
                axes=axes[-1],
                outlines=outlines,
            )
            if idx == n_blocs - 1:
                axes.append(fig.add_subplot(grid[i, n_blocs * 3]))
                fig.colorbar(
                    im,
                    ax=axes[-1],
                    location="right",
                    shrink=0.9,
                    ticks=(-vlim, 0, vlim),
                )
                axes[-1].axis("off")

    out_path = os.path.join(save_path, f"{suffix}saliencies.png")
    # plt.tight_layout()
    plt.savefig(out_path, dpi=300)
    if show:
        plt.show()
    plt.close()
    return out_path


def colorFader(c1, c2, mix=0, alpha=0.5):
    # fade (linear interpolate) from color c1 (at mix=0) to c2 (mix=1)
    c1 = np.array(mpl.colors.to_rgb(c1))
    c2 = np.array(mpl.colors.to_rgb(c2))
    return mpl.colors.to_hex((1 - mix) * c1 + mix * c2)


def plot_masked_epoch(data, mask, c1="white", c2="green", alpha=0.3, title: str = None):
    dim = len(data.shape)
    assert data.shape == mask.shape, "Shape of the data should match shape of the mask."
    assert (
        dim == 2
    ), f"Data is supposed to be a 2 dimension array, a {dim}-dimension array was given."

    mask = (mask - mask.min()) / np.ptp(mask)
    fig, axes = plot_epoch(data, title)
    for i, ax in enumerate(axes):
        for j, x in enumerate(range(data.shape[1])):
            ax.axvline(
                x + 1,
                color=colorFader(c1, c2, mask[i, j]),
                zorder=0,
                alpha=alpha,
                linewidth=50,
            )

    return fig, axes


def plot_epoch(data, title: str = None):
    dim = len(data.shape)
    assert (
        dim == 2
    ), f"Data is supposed to be a 2 dimension array, a {dim}-dimension array was given."

    grid = GridSpec(len(data), 1, hspace=0)
    fig = plt.figure(figsize=(10, 5))
    plt.title(title)
    plt.axis("off")
    axes = []
    for i, chan in enumerate(data):
        ax = fig.add_subplot(grid[i, 0])
        plt.plot(chan, color="black", linewidth=1)
        plt.xlim(0, data.shape[1])
        ax.spines["top"].set_visible(False)
        ax.spines["right"].set_visible(False)
        # plt.ylabel(i, rotation="horizontal")
        ax.spines["left"].set_visible(False)
        ax.set_yticks([])
        if i != len(data) - 1:
            ax.spines["bottom"].set_visible(False)
            ax.set_xticks([])
        axes.append(ax)
    plt.tight_layout()

    # out_path = os.path.join(save_path, "visualizations", "average_trial.png")
    # plt.savefig(out_path, dpi=300)
    # return out_path
    return fig, axes


def compute_cams(net, target_layers, dataset, verbose=3):
    assert (
        len(dataset.subject_list) != 0
    ), "No subject in the dataset, please use dataset.preload before using this function."
    assert len(dataset.data) == 0, "Do not load the dataset before using this function."
    all_cams = [[], []]

    for sub in dataset.subject_list:
        sub_dataset = copy.deepcopy(dataset)
        if verbose > 2:
            print(f"Processing subject {sub}")
        sub_dataset.load(dataset.data_path, one_sub=sub, verbose=verbose)
        if len(sub_dataset) == 0:
            continue

        for i, _ in enumerate(sub_dataset.target_labels):
            gradcam = GradCAM(model=net, target_layers=target_layers)
            input_tensor = sub_dataset.data[sub_dataset.targets == i]

            # Check model confidence
            outputs = net(input_tensor.to("cuda"))  # Forward pass
            probabilities = F.softmax(outputs, dim=1)
            confidences = probabilities[:, i].cpu()  # Confidence for class `i`
            high_conf_indices = confidences > 0.9  # Filter high-confidence predictions

            if (
                high_conf_indices.sum() > 0
            ):  # Only proceed if there are high-confidence predictions
                high_conf_input_tensor = input_tensor[high_conf_indices]

                # Compute Grad-CAM only for high-confidence samples
                grayscale_cam = gradcam(
                    input_tensor=high_conf_input_tensor, targets=[ClassifierOutputTarget(i)]
                )
                valid_cams = [
                    cam for cam in grayscale_cam if not np.isnan(cam).any() and cam.max() > 0
                ]

                if len(valid_cams) > 0:
                    average_cam = np.mean(valid_cams, axis=0)
                    all_cams[i].append(average_cam)

    return [np.array(label_cams) for label_cams in all_cams]


def plot_cam(input_tensor, cams, name, label, out_path=".", colorbar=True):
    cam_output_path = os.path.join(out_path, f"{name}_{label}_GradCAM.png")

    grayscale_cam = cams.mean(axis=0)
    grayscale_cam = (grayscale_cam - grayscale_cam.min()) / (
        grayscale_cam.max() - grayscale_cam.min()
    )
    img = np.moveaxis(np.array(input_tensor[0]), 0, 2)
    img = np.array([img[:, :, 0]] * 3)
    img = np.moveaxis(img, 0, 2)
    cam_image = show_cam_on_image(img, grayscale_cam, use_rgb=True)
    cam_image = cv2.cvtColor(cam_image, cv2.COLOR_RGB2BGR)

    method = "GradCAM"

    if os.path.exists(cam_output_path):
        os.remove(cam_output_path)
    cv2.imwrite(cam_output_path, cam_image)

    # Plot with colorbar and axis labels
    fig, ax = plt.subplots()
    ax.imshow(img, aspect="auto")  # Show the original image
    im = ax.imshow(grayscale_cam, cmap="jet", alpha=0.7)  # Overlay heatmap
    ax.set_title(f"{name}_{method}_{label} Grad-CAM")
    ax.set_xlabel("Time (s)")
    ax.set_ylabel("Sensors")
    ax.set_xticks([0, 75, 150, 225, 300, 375])  # Set x-axis ticks
    ax.set_xticklabels([-0.15, 0, 0.15, 0.3, 0.45, 0.6])
    ax.set_yticks([])  # Remove y-axis ticks

    if colorbar:
        divider = make_axes_locatable(ax)
        cax = divider.append_axes("right", size="5%", pad=0.15)
        cbar = plt.colorbar(im, cax=cax)
        cbar.set_ticks([0, 0.5, 1])
        cbar.set_ticklabels([0, 0.5, 1])

    plt.savefig(os.path.join(out_path, f"{name}_{label}_GradCAM_with_colorbar.png"))
    plt.show()
    return fig
